# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import mando
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils


@mando.command("pca", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def pca_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    n_components=None,
    source_units=None,
    target_units=None,
    round_index=None,
    tablefmt="csv",
):
    """Return the principal components analysis of the time series.

    Does not return a time-series.

    Parameters
    ----------
    n_components : int
        [optional, default is None]

        The columns in the input_ts will be grouped into `n_components`
        groups.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${clean}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        pca(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            clean=clean,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            n_components=n_components,
            source_units=source_units,
            target_units=target_units,
            round_index=round_index,
        ),
        tablefmt=tablefmt,
    )


@typic.al
@tsutils.copy_doc(pca_cli)
def pca(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    n_components: tsutils.IntGreaterEqualToOne = None,
    source_units=None,
    target_units=None,
    round_index=None,
):
    """Return the principal components analysis of the time series."""
    from sklearn.decomposition import PCA

    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        round_index=round_index,
        pick=columns,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    pca = PCA(n_components)
    pca.fit(tsd.dropna(how="any"))
    return pca.components_



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import re
import warnings
from typing import List

import mando
import numpy as np
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter
from numpy import *  # nopycln: import

from .. import tsutils

warnings.filterwarnings("ignore")


def _parse_equation(equation_str):
    """Private function to parse the equation used in the calculations."""
    # Get rid of spaces
    nequation = equation_str.replace(" ", "")

    # Does the equation contain any x[t]?
    tsearch = re.search(r"\[.*?t.*?\]", nequation)

    # Does the equation contain any x1, x2, ...etc.?
    nsearch = re.search(r"x[1-9][0-9]*", nequation)

    # This beasty is so users can use 't' in their equations
    # Indices of 'x' are a function of 't' and can possibly be negative or
    # greater than the length of the DataFrame.
    # Correctly (I think) handles negative indices and indices greater
    # than the length by setting to nan
    # AssertionError happens when index negative.
    # IndexError happens when index is greater than the length of the
    # DataFrame.
    # UGLY!

    # testeval is just a list of the 't' expressions in the equation.
    # for example 'x[t]+0.6*max(x[t+1],x[t-1]' would have
    # testeval = ['t', 't+1', 't-1']
    testeval = set()

    # If there is both function of t and column terms x1, x2, ...etc.
    if tsearch and nsearch:
        testeval.update(re.findall(r"x[1-9][0-9]*\[(.*?t.*?)\]", nequation))

        # replace 'x1[t+1]' with 'x.iloc[t+1,1-1]'
        # replace 'x2[t+7]' with 'x.iloc[t+7,2-1]'
        # ...etc
        nequation = re.sub(
            r"x([1-9][0-9]*)\[(.*?t.*?)\]", r"x.iloc[\2,\1-1]", nequation
        )

        # replace 'x1' with 'x.iloc[t,1-1]'
        # replace 'x4' with 'x.iloc[t,4-1]'
        # ...etc
        nequation = re.sub(r"x([1-9][0-9]*)", r"x.iloc[t,\1-1]", nequation)

    # If there is only a function of t, i.e. x[t]
    elif tsearch:
        testeval.update(re.findall(r"x\[(.*?t.*?)\]", nequation))
        nequation = re.sub(r"x\[(.*?t.*?)\]", r"xxiloc[\1,:]", nequation)
        # Replace 'x' with underlying equation, but not the 'x' in a word,
        # like 'maximum'.
        nequation = re.sub(r"(?<![a-zA-Z])x(?![a-zA-Z\[])", r"xxiloc[t,:]", nequation)
        nequation = re.sub(r"xxiloc", r"x.iloc", nequation)

    elif nsearch:
        nequation = re.sub(r"x([1-9][0-9]*)", r"x.iloc[:,\1-1]", nequation)

    try:
        testeval.remove("t")
    except KeyError:
        pass
    return tsearch, nsearch, testeval, nequation


@mando.command("equation", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def equation_cli(
    equation_str,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    print_input="",
    round_index=None,
    source_units=None,
    target_units=None,
    float_format="g",
    tablefmt="csv",
    output_names="",
):
    """Apply <equation_str> cto the time series data.

    The <equation_str> argument is a string contained in single quotes
    with 'x', 'x[t]', or 'x1', 'x2', 'x3', ...etc. used as the variable
    representing the input.  For example, '(1 - x)*sin(x)'.

    Parameters
    ----------
    equation_str : str
        String contained in single quotes that defines the equation.

        Can have multiple equations separated by an "@" symbol.

        There are four different types of equations that can be used.

        +-----------------------+-----------+-------------------------+
        | Description           | Variables | Examples                |
        +=======================+===========+=========================+
        | Equation applied to   | x         | x*0.3+4-x**2            |
        | all values in the     |           | sin(x)+pi*x             |
        | dataset.  Returns     |           |                         |
        | same number of        |           |                         |
        | columns as input.     |           |                         |
        +-----------------------+-----------+-------------------------+
        | Equation used time    | x and t   | 0.6*max(x[t-1],x[t+1])  |
        | relative to current   |           |                         |
        | record.  Applies      |           |                         |
        | equation to each      |           |                         |
        | column.  Returns same |           |                         |
        | number of columns as  |           |                         |
        | input.                |           |                         |
        +-----------------------+-----------+-------------------------+
        | Equation uses values  | x1, x2,   | x1+x2                   |
        | from different        | x3, ...   |                         |
        | columns.  Returns a   | xN        |                         |
        | single column.        |           |                         |
        +-----------------------+-----------+-------------------------+
        | Equation uses values  | x1, x2,   | x1[t-1]+x2+x3[t+1]      |
        | from different        | x3,       |                         |
        | columns and values    | ...xN, t  |                         |
        | from different rows.  |           |                         |
        | Returns a single      |           |                         |
        | column.               |           |                         |
        +-----------------------+-----------+-------------------------+

        Mathematical functions in the 'np' (numpy) name space can be
        used.  Additional examples::

            'x*4 + 2',
            'x**2 + cos(x)', and
            'tan(x*pi/180)'

        are all valid <equation> strings.  The variable 't' is special
        representing the index (usually time) at which 'x' occurs.  This
        means you can do things like::

            'x[t] + max(x[t-1], x[t+1])*0.6'

        to add to the current value 0.6 times the maximum row adjacent
        value.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${clean}
    ${print_input}
    ${names}
    ${float_format}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    ${output_names}
    """
    tsutils.printiso(
        equation(
            equation_str,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            print_input=print_input,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
            output_names=output_names,
        ),
        float_format=float_format,
        tablefmt=tablefmt,
    )


@tsutils.transform_args(output_names=tsutils.make_list)
@typic.al
@tsutils.copy_doc(equation_cli)
def equation(
    equation_str: str,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    print_input="",
    round_index=None,
    source_units=None,
    target_units=None,
    output_names: List[str] = None,
):
    """Apply <equation_str> to the time series data."""
    if output_names is None:
        output_names = []
    x = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    def returnval(t, x, testeval, nequation):
        """Return the 'eval'uated equation.

        Need "x" and "t" as those variables since will be used in "equation"
        even if they do not directly appear in the "returnval" function.
        """
        for tst in testeval:
            # If any of the equations in "testeval" are < 0 or > len(x) must be
            # np.nan.
            tvalue = eval(tst)
            if tvalue < 0 or tvalue >= len(x):
                return np.nan
        return eval(nequation)

    newy = pd.DataFrame()
    for cnt, eqn in enumerate(tsutils.make_list(equation_str, sep="@")):
        tsearch, nsearch, testeval, nequation = _parse_equation(eqn)
        if tsearch and nsearch:
            y = pd.DataFrame(
                pd.Series(index=x.index, dtype=float),
                columns=["_"],
                dtype=float,
            )
            for t in range(len(x)):
                y.iloc[t, 0] = returnval(t, x, testeval, nequation)
        elif tsearch:
            y = x.astype(float)
            for t in range(len(x)):
                y.iloc[t, :] = returnval(t, x, testeval, nequation)
        elif nsearch:
            y = pd.DataFrame(
                pd.Series(index=x.index, dtype=float),
                columns=["_"],
                dtype=float,
            )
            try:
                y.iloc[:, 0] = eval(nequation)
            except IndexError:
                raise IndexError(
                    """
*
*   There are {} columns, but the equation you are trying to apply is trying
*   to access a column greater than that.
*
""".format(
                        y.shape[1]
                    )
                )

        else:
            y = eval(eqn)
        y.columns = [tsutils.renamer(i, "equation{}_".format(cnt)) for i in y.columns]
        newy = newy.join(y, how="outer")

    newy = tsutils.memory_optimize(newy)

    return tsutils.return_input(print_input, x, newy, output_names=output_names)



# -*- coding: utf-8 -*-

import os
import tempfile
from unittest import TestCase

import numpy as np
import pandas as pd

from tstoolbox import tstoolbox

pandacodes = [
    "A",
    "AS",
    "BA",
    "BAS",  # Annual
    "Q",
    "QS",
    "BQ",
    "BQS",  # Quarterly
    "M",
    "MS",
    "BM",
    "BMS",  # Monthly
    "W",  # Weekly
    "D",
    "B",  # Daily
    "H",
    "T",
    "S",
    "L",
    "U",
]  # Intra-daily

pd_tstep_minterval = {
    "A": ("1800-01-01", 300, 10),
    "AS": ("1800-01-01", 300, 10),
    "BA": ("1800-01-01", 300, 10),
    "BAS": ("1800-01-01", 300, 10),
    "Q": ("1900-01-01", 300, 12),
    "QS": ("1900-01-01", 300, 12),
    "BQ": ("1900-01-01", 300, 12),
    "BQS": ("1900-01-01", 300, 12),
    "M": ("1900-01-01", 240, 12),
    "MS": ("1900-01-01", 240, 12),
    "BM": ("1900-01-01", 240, 12),
    "BMS": ("1900-01-01", 240, 12),
    "W": ("2000-01-01", 240, 12),
    "D": ("2000-01-01", 3660, 10),
    "B": ("2000-01-03", 3660, 10),
    "H": ("2000-01-01", 24 * 366 * 5, 5),
    "T": ("2000-01-01", 60 * 24 * 366, 2),
    "S": ("2000-01-01", 60 * 60 * 24 * 2, 3),
    "L": ("2000-01-01", 1000, 2),
    "U": ("2000-01-01", 1000, 2),
}

import logging

logging.basicConfig(filename="example.log", encoding="utf-8", level=logging.DEBUG)


class TestAddTrend(TestCase):
    def setUp(self):
        """Create a whole bunch of time-series at different intervals.

        Uses 'pandacodes' and different number of intervals, the maximum number
        of intervals is the third term in each entry in pd_tstep_minterval.
        Then writes them out to files that will be read in the tests.
        """
        self.fps = {}
        for testpc in pandacodes:
            sdate, periods, nintervals = pd_tstep_minterval[testpc]
            for tstep in range(1, nintervals):
                aperiods = periods // tstep
                dr = pd.date_range(
                    sdate, periods=aperiods, freq="{}{}".format(tstep, testpc)
                )
                df = pd.DataFrame(np.arange(aperiods), index=dr)
                df.index.name = "index"
                df.columns = ["data"]
                self.fps[(tstep, testpc)] = tempfile.mkstemp()
                df.to_csv(self.fps[(tstep, testpc)][1], sep=",", header=True)

    def test_ts_intervals(self):
        """Test many intervals to make sure tstoolbox makes a good guess."""
        # 'matches' lists out things that should match but tstoolbox will give
        # the simpler answer.  For example 4 quarters is equal to a 1 year
        # interval.  Also difficult for tstoolbox to figure out that a 2 month
        # interval isn't a 1 month interval with gaps.
        matches = {
            "4Q": "1A",
            "8Q": "2A",
            "4BQS": "1BAS",
            "8BQS": "2BAS",
            "4BQ": "1BA",
            "8BQ": "2BA",
            "4QS": "1AS",
            "8QS": "2AS",
            "2M": "1M",
            "3M": "1Q",
            "4M": "1M",
            "5M": "1M",
            "6M": "2Q",
            "7M": "1M",
            "8M": "1M",
            "9M": "3Q",
            "10M": "1M",
            "11M": "1M",
            "2BMS": "1BMS",
            "3BMS": "1BQS",
            "4BMS": "1BMS",
            "5BMS": "1BMS",
            "6BMS": "2BQS",
            "7BMS": "1BMS",
            "8BMS": "1BMS",
            "9BMS": "3BQS",
            "10BMS": "1BMS",
            "11BMS": "1BMS",
            "2BM": "1BM",
            "3BM": "1BQ",
            "4BM": "1BM",
            "5BM": "1BM",
            "6BM": "2BQ",
            "7BM": "1BM",
            "8BM": "1BM",
            "9BM": "3BQ",
            "10BM": "1BM",
            "11BM": "1BM",
            "2MS": "1MS",
            "3MS": "1QS",
            "4MS": "1MS",
            "5MS": "1MS",
            "6MS": "2QS",
            "7MS": "1MS",
            "8MS": "1MS",
            "9MS": "3QS",
            "10MS": "1MS",
            "11MS": "1MS",
            "5B": "1W",
            "7D": "1W",
            # BUG!!!
            "2B": "2D",
            "3B": "1D",
            "4B": "2D",
            "6B": "2D",
            "7B": "1D",
            "8B": "2D",
            "9B": "1D",
        }

        for key in self.fps:
            df = tstoolbox.read(self.fps[key][1])
            inferred_code = df.index.inferred_freq
            if inferred_code is None:
                continue
            testcode = "{}{}".format(*key)
            if inferred_code[0] not in "123456789":
                inferred_code = "1" + inferred_code
            testcode = matches.get(testcode, testcode)
            logging.warning("{} {}".format(testcode, inferred_code))
            try:
                self.assertEqual(testcode, inferred_code.split("-")[0])
            except AssertionError:
                self.assertEqual(
                    [i for i in testcode if not i.isdigit()],
                    [i for i in inferred_code.split("-")[0] if not i.isdigit()],
                )

    def tearDown(self):
        """Remove the temporary files."""
        for key, fname in self.fps.items():
            if os.path.exists(fname[1]):
                os.remove(fname[1])



# -*- coding: utf-8 -*-

from __future__ import print_function

from unittest import TestCase

import pandas as pd
import pytest
from pandas.testing import assert_index_equal

from tstoolbox import tstoolbox, tsutils


class TestRead(TestCase):
    def setUp(self):
        """Prepare in-memory versions of the files ./data_flat.csv."""
        self.data = pd.read_csv(
            "tests/data_flat.csv", sep=",", parse_dates=True, index_col=0
        )
        self.data.index.name = "Datetime"

    def test_createts_from_input(self):
        """Create a ts of data_flat.csv."""
        out = tstoolbox.createts(input_ts=self.data).index
        out1 = tstoolbox.read("tests/data_flat.csv").index
        assert_index_equal(out, out1)

    def test_createts_from_dates(self):
        """Create a ts file from start/end dates and freq."""
        sdate = self.data.index[0]
        edate = self.data.index[-1]
        freq = tsutils.asbestfreq(self.data)
        freq = freq.index.freqstr
        out = tstoolbox.createts(start_date=sdate, end_date=edate, freq=freq).index
        out.name = "Datetime"
        assert_index_equal(out, self.data.index)

    @staticmethod
    def test_exception():
        with pytest.raises(ValueError) as e_info:
            _ = tstoolbox.createts()
        assert r"If input_ts is None, then start_date, end_date" in str(e_info.value)



# -*- coding: utf-8 -*-
from unittest import TestCase

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

import tstoolbox.tsutils as tsutils
from tstoolbox import tstoolbox

from . import capture


class TestReplace(TestCase):
    def setUp(self):
        dindex = pd.date_range("2011-01-01T00:00:00", periods=26, freq="H")
        self.ats = np.ones(26)
        self.ats = pd.DataFrame(
            self.ats, index=dindex, columns=["Value_with_missing::replace"]
        )
        self.ats.index.name = "Datetime"

        self.ats_cli = capture.capture(tsutils.printiso, self.ats)

        self.freplace_compare = self.ats.copy()
        self.freplace_compare["Value_with_missing::replace"][
            "2011-01-01T09:00:00":"2011-01-01T12:00:00"
        ] = 3
        self.freplace_compare["Value_with_missing::replace"]["2011-01-01T13:00:00"] = 9

        self.freplace_compare.columns = ["Value_with_missing"]
        self.freplace_compare_cli = capture.capture(
            tsutils.printiso, self.freplace_compare
        )

        self.breplace_compare = self.ats.copy()
        self.breplace_compare["Value_with_missing::replace"]["2011-01-01T09:00:00"] = 3
        self.breplace_compare["Value_with_missing::replace"][
            "2011-01-01T10:00:00":"2011-01-01T13:00:00"
        ] = 9

        self.breplace_compare.columns = ["Value_with_missing"]
        self.breplace_compare_cli = capture.capture(
            tsutils.printiso, self.breplace_compare
        )

        self.linear_compare = self.ats.copy()
        self.linear_compare["Value_with_missing::replace"]["2011-01-01T09:00:00"] = 3.0
        self.linear_compare["Value_with_missing::replace"]["2011-01-01T10:00:00"] = 4.5
        self.linear_compare["Value_with_missing::replace"]["2011-01-01T11:00:00"] = 6.0
        self.linear_compare["Value_with_missing::replace"]["2011-01-01T12:00:00"] = 7.5
        self.linear_compare["Value_with_missing::replace"]["2011-01-01T13:00:00"] = 9.0

        self.linear_compare.columns = ["Value_with_missing"]
        self.linear_compare_cli = capture.capture(tsutils.printiso, self.linear_compare)

        self.nearest_compare = self.ats.copy()
        self.nearest_compare["Value_with_missing::replace"][
            "2011-01-01T09:00:00":"2011-01-01T11:00:00"
        ] = 3.0
        self.nearest_compare["Value_with_missing::replace"][
            "2011-01-01T12:00:00":"2011-01-01T13:00:00"
        ] = 9.0

        self.nearest_compare.columns = ["Value_with_missing"]
        self.nearest_compare_cli = capture.capture(
            tsutils.printiso, self.nearest_compare
        )

        self.median_compare = self.ats.copy()
        self.median_compare["Value_with_missing::replace"]["2011-01-01T01:00:00"] = 2.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T09:00:00"] = 3.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T10:00:00"] = 2.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T11:00:00"] = 2.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T12:00:00"] = 2.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T13:00:00"] = 9.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T16:00:00"] = 2.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T22:00:00"] = 2.0
        self.median_compare["Value_with_missing::replace"]["2011-01-01T23:00:00"] = 2.0

        self.median_compare.columns = ["Value_with_missing"]
        self.median_compare_cli = capture.capture(tsutils.printiso, self.median_compare)

        self.max_compare = self.ats.copy()
        self.max_compare["Value_with_missing::replace"]["2011-01-01T01:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T09:00:00"] = 3.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T10:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T11:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T12:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T13:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T16:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T22:00:00"] = 9.0
        self.max_compare["Value_with_missing::replace"]["2011-01-01T23:00:00"] = 9.0

        self.max_compare.columns = ["Value_with_missing"]
        self.max_compare_cli = capture.capture(tsutils.printiso, self.max_compare)

        self.min_compare = self.ats.copy()
        self.min_compare["Value_with_missing::replace"]["2011-01-01T01:00:00"] = 2.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T09:00:00"] = 3.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T10:00:00"] = 2.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T11:00:00"] = 2.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T12:00:00"] = 2.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T13:00:00"] = 9.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T16:00:00"] = 2.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T22:00:00"] = 2.0
        self.min_compare["Value_with_missing::replace"]["2011-01-01T23:00:00"] = 2.0

        self.min_compare.columns = ["Value_with_missing"]
        self.min_compare_cli = capture.capture(tsutils.printiso, self.min_compare)

        self.con_compare = self.ats.copy()
        self.con_compare["Value_with_missing::replace"]["2011-01-01T01:00:00"] = 2.42
        self.con_compare["Value_with_missing::replace"]["2011-01-01T09:00:00"] = 3.0
        self.con_compare["Value_with_missing::replace"]["2011-01-01T10:00:00"] = 2.42
        self.con_compare["Value_with_missing::replace"]["2011-01-01T11:00:00"] = 2.42
        self.con_compare["Value_with_missing::replace"]["2011-01-01T12:00:00"] = 2.42
        self.con_compare["Value_with_missing::replace"]["2011-01-01T13:00:00"] = 9.0
        self.con_compare["Value_with_missing::replace"]["2011-01-01T16:00:00"] = 2.42
        self.con_compare["Value_with_missing::replace"]["2011-01-01T22:00:00"] = 2.42
        self.con_compare["Value_with_missing::replace"]["2011-01-01T23:00:00"] = 2.42

        self.con_compare.columns = ["Value_with_missing"]
        self.con_compare_cli = capture.capture(tsutils.printiso, self.con_compare)

    def test_replace_freplace_direct(self):
        """Test forward replace API."""
        out = tstoolbox.replace([3, 9], [1, 1], input_ts=self.freplace_compare)
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_breplace(self):
        """Test backward replace API."""
        out = tstoolbox.replace([3, 9], [1, 1], input_ts=self.breplace_compare)
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_linear(self):
        """Test linear interpolation replace API."""
        out = tstoolbox.replace(
            [3.0, 4.5, 6.0, 7.5, 9.0], [1, 1, 1, 1, 1], input_ts=self.linear_compare
        )
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_nearest(self):
        """Test nearest replace API."""
        out = tstoolbox.replace([3.0, 9.0], [1, 1], input_ts=self.nearest_compare)
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_median(self):
        """Test replace with median API."""
        out = tstoolbox.replace([2, 3, 9], [1, 1, 1], input_ts=self.median_compare)
        self.maxDiff = None
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_max(self):
        """Test replace with max API."""
        out = tstoolbox.replace([3, 9], [1, 1], input_ts=self.max_compare)
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_min(self):
        """Test replace with min API."""
        out = tstoolbox.replace([2, 3, 9], [1, 1, 1], input_ts=self.min_compare)
        assert_frame_equal(out, self.ats, check_dtype=False)

    def test_replace_con(self):
        """Test replace with con API."""
        out = tstoolbox.replace([2.42, 3, 9], [1, 1, 1], input_ts=self.con_compare)
        assert_frame_equal(out, self.ats, check_dtype=False)


#    def test_replace_freplace_cli(self):
#        """Test forward replace CLI."""
#        args = 'tstoolbox replace --input_ts=tests/data_missing.csv'
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#                               stdout=subprocess.PIPE,
#                               stdin=subprocess.PIPE).communicate(input=self.ats_cli)[0]
#        self.maxDiff = None
#        self.assertEqual(out, self.freplace_compare_cli)
#
#    def test_replace_breplace_cli(self):
#        """Test backward replace CLI."""
#        args = ('tstoolbox replace '
#                '--method="breplace" '
#                '--input_ts=tests/data_missing.csv')
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#                               stdout=subprocess.PIPE,
#                               stdin=subprocess.PIPE).communicate(input=self.ats_cli)[0]
#        self.maxDiff = None
#        self.assertEqual(out, self.breplace_compare_cli)
#
#    def test_replace_linear_cli(self):
#        """Test linear replace CLI."""
#        args = ('tstoolbox replace '
#                '--method="linear" '
#                '--input_ts=tests/data_missing.csv')
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#                               stdout=subprocess.PIPE,
#                               stdin=subprocess.PIPE).communicate(input=self.ats_cli)[0]
#        self.maxDiff = None
#        self.assertEqual(out, self.linear_compare_cli)
#
#    def test_replace_nearest_cli(self):
#        """Test nearest replace CLI."""
#        args = ('tstoolbox replace '
#                '--method="nearest" '
#                '--input_ts=tests/data_missing.csv')
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#                               stdout=subprocess.PIPE,
#                               stdin=subprocess.PIPE).communicate(input=self.ats_cli)[0]
#        self.maxDiff = None
#        self.assertEqual(out, self.nearest_compare_cli)
#
#    def test_replace_mean_cli(self):
#        """Test mean replace CLI."""
#        args = ('tstoolbox replace '
#                '--method="mean" '
#                '--input_ts=tests/data_missing.csv')
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#                               stdout=subprocess.PIPE,
#                               stdin=subprocess.PIPE).communicate(input=self.ats_cli)[0]
#        self.maxDiff = None
#        self.assertEqual(out, self.mean_compare_cli)



# -*- coding: utf-8 -*-
import numpy as np

from . import utils


def rmsd(predicted, reference):
    """
    Calculate root-mean-square deviation (RMSD) between two variables.

    Calculates the root-mean-square deviation between two variables
    PREDICTED and REFERENCE. The RMSD is calculated using the
    formula:

    RMSD^2 = sum_(n=1)^N [(p_n - r_n)^2]/N

    where p is the predicted values, r is the reference values, and
    N is the total number of values in p & r. Note that p & r must
    have the same number of values.

    Input:
    PREDICTED : predicted values
    REFERENCE : reference values

    Output:
    R : root-mean-square deviation (RMSD)

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 9, 2016
    """
    # Check that dimensions of predicted and reference fields match
    utils.check_arrays(predicted, reference)

    # Calculate the RMSE
    r = np.sqrt(np.sum(np.square(predicted - reference)) / len(predicted))

    return r



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import os
import warnings
from argparse import RawTextHelpFormatter

import mando
import pandas as pd
import typic

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@mando.command("read", formatter_class=RawTextHelpFormatter)
@tsutils.doc(tsutils.docstrings)
def read_cli(
    force_freq=None,
    append="columns",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    source_units=None,
    target_units=None,
    float_format="g",
    round_index=None,
    tablefmt="csv",
    *filenames,
):
    """Combine time-series from different sources into single dataset.

    Prints the read in time-series in the tstoolbox standard format.

    WARNING: Accepts naive and timezone aware time-series by converting all to
    UTC and removing timezone information.

    Parameters
    ----------
    filenames : str
        From the command line a list of comma or space delimited filenames to
        read time series from.  Using the Python API a list or tuple of
        filenames.
    append : str
        [optional, default is 'columns']

        The type of appending to do.  For "combine" option matching column
        indices will append rows, matching row indices will append columns, and
        matching column/row indices use the value from the first dataset.  You
        can use "row" or "column" to force an append along either axis.
    ${force_freq}
        ${pandas_offset_codes}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${float_format}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        read(
            *filenames,
            append=append,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            index_type=index_type,
            clean=clean,
            force_freq=force_freq,
            round_index=round_index,
            columns=columns,
            skiprows=skiprows,
            names=names,
            source_units=source_units,
            target_units=target_units,
        ),
        float_format=float_format,
        tablefmt=tablefmt,
    )


@typic.al
@tsutils.copy_doc(read_cli)
def read(
    *filenames,
    force_freq=None,
    append: Literal["columns", "rows", "combine"] = "columns",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    source_units=None,
    target_units=None,
    round_index=None,
):
    """Collect time series from a list of pickle or csv files."""
    if force_freq is not None:
        dropna = "no"

    if isinstance(filenames, (list, tuple)) and len(filenames) == 1:
        filenames = filenames[0]

    # Check for older style where comma delimited list of only files.
    # If so, rework as space delimited.
    isspacedelimited = False
    for fname in tsutils.make_list(filenames, sep=","):
        if not os.path.exists(str(fname)):
            isspacedelimited = True
            break

    if isspacedelimited is True:
        filenames = tsutils.make_list(filenames, sep=" ", flat=False)
    else:
        # All filenames are real files.  Therefore old style and just make
        # a simple list.
        filenames = tsutils.make_list(filenames, sep=",")
        warnings.warn(
            tsutils.error_wrapper(
                """
Using "," separated files is deprecated in favor of space delimited files."""
            )
        )

    tsd = tsutils.common_kwds(
        input_tsd=filenames,
        skiprows=skiprows,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        round_index=round_index,
        names=names,
        dropna=dropna,
        force_freq=force_freq,
        clean=clean,
        source_units=source_units,
        target_units=target_units,
        usecols=columns,
    )

    return tsd



# -*- coding: utf-8 -*-
from unittest import TestCase

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class TestRead(TestCase):
    def setUp(self):
        base = pd.read_csv(
            "tests/data_missing.csv",
            index_col=[0],
            parse_dates=True,
            skipinitialspace=True,
        ).astype("Int64")
        base.index.name = "Datetime"
        self.cumsum = base.cumsum()
        self.cumsum.columns = [tsutils.renamer(i, "sum") for i in self.cumsum.columns]

    def test_cumsum(self):
        """Test cumsum."""
        out = tstoolbox.accumulate(input_ts="tests/data_missing.csv", dropna="any")
        assert_frame_equal(out, self.cumsum, check_dtype=False)


def test_stats():
    """Test stat names."""
    with pytest.raises(ValueError):
        _ = tstoolbox.accumulate(
            input_ts="tests/data_missing.csv", dropna="any", statistic="camel"
        )



# -*- coding: utf-8 -*-
import numpy as np

from . import utils


def pc_bias(simulated, observed):
    """
    Calculate the percent bias between simulated and observed.

    B = 100.0*sum(s-o)/sum(o)

    where s is the simulated values, and o is the observed values.
    Note that s & o must have the same number of values.

    Input:
    simulated : simulated field
    observed : observed field

    Output:
    B : percent bias between simulated and observed
    """
    # Check that dimensions of simulated and observed fields match
    utils.check_arrays(simulated, observed)

    # Calculate bias in means
    b = 100.0 * np.sum(simulated - observed) / np.sum(observed)

    return b



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings
from typing import Union

import mando
import numpy as np
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter
from pandas.tseries.frequencies import to_offset

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


@mando.command("convert_index", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def convert_index_cli(
    to,
    interval=None,
    epoch="julian",
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    clean=False,
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
    tablefmt="csv",
):
    """Convert datetime to/from Julian dates from different epochs.

    Parameters
    ----------
    to: str
        One of 'number' or 'datetime'.  If 'number', the source time-series
        should have a datetime index to convert to a number.  If 'datetime',
        source data should be a number and the converted index will be
        datetime.
    interval
        [optional, defaults to None, transformation]

        The `interval` parameter defines the unit time.  One of the pandas
        offset codes.  The default of 'None' will set the unit time for all
        defined epochs to daily except 'unix' which will default to seconds.

        You can give any smaller unit time than daily for all defined epochs
        except 'unix' which requires an interval less than seconds.  For an
        epoch that begins with an arbitrary date, you can use any interval
        equal to or smaller than the frequency of the time-series.

        ${pandas_offset_codes}
    epoch : str
        [optional, defaults to 'julian', transformation]

        Can be one of, 'julian', 'reduced', 'modified', 'truncated', 'dublin',
        'cnes', 'ccsds', 'lop', 'lilian', 'rata_die', 'mars_sol_date', 'unix',
        or a date and time.

        If supplying a date and time, most formats are recognized, however
        the closer the format is to ISO 8601 the better.  Also should check and
        make sure date was parsed as expected.  If supplying only a date, the
        epoch starts at midnight the morning of that date.

        The 'unix' epoch uses a default `interval` of seconds, and all other
        defined epochs use a default `interval` of 'daily'.

        +-----------+----------------+----------------+-------------+
        | epoch     | Epoch          | Calculation    | Notes       |
        +===========+================+================+=============+
        | julian    | 4713-01-01:12  | JD             |             |
        |           | BCE            |                |             |
        +-----------+----------------+----------------+-------------+
        | reduced   | 1858-11-16:12  | JD -           | [ [1]_ ]    |
        |           |                | 2400000        | [ [2]_ ]    |
        +-----------+----------------+----------------+-------------+
        | modified  | 1858-11-17:00  | JD -           | SAO 1957    |
        |           |                | 2400000.5      |             |
        +-----------+----------------+----------------+-------------+
        | truncated | 1968-05-24:00  | floor (JD -    | NASA 1979,  |
        |           |                | 2440000.5)     | integer     |
        +-----------+----------------+----------------+-------------+
        | dublin    | 1899-12-31:12  | JD -           | IAU 1955    |
        |           |                | 2415020        |             |
        +-----------+----------------+----------------+-------------+
        | cnes      | 1950-01-01:00  | JD -           | CNES        |
        |           |                | 2433282.5      | [ [3]_ ]    |
        +-----------+----------------+----------------+-------------+
        | ccsds     | 1958-01-01:00  | JD -           | CCSDS       |
        |           |                | 2436204.5      | [ [3]_ ]    |
        +-----------+----------------+----------------+-------------+
        | lop       | 1992-01-01:00  | JD -           | LOP         |
        |           |                | 2448622.5      | [ [3]_ ]    |
        +-----------+----------------+----------------+-------------+
        | lilian    | 1582-10-15[13] | floor (JD -    | Count of    |
        |           |                | 2299159.5)     | days of the |
        |           |                |                | Gregorian   |
        |           |                |                | calendar,   |
        |           |                |                | integer     |
        +-----------+----------------+----------------+-------------+
        | rata_die  | 0001-01-01[13] | floor (JD -    | Count of    |
        |           | proleptic      | 1721424.5)     | days of the |
        |           | Gregorian      |                | Common      |
        |           | calendar       |                | Era,        |
        |           |                |                | integer     |
        +-----------+----------------+----------------+-------------+
        | mars_sol  | 1873-12-29:12  | (JD - 2405522) | Count of    |
        |           |                | /1.02749       | Martian     |
        |           |                |                | days        |
        +-----------+----------------+----------------+-------------+
        | unix      | 1970-01-01     | JD - 2440587.5 | seconds     |
        |           | T00:00:00      |                |             |
        +-----------+----------------+----------------+-------------+

        .. [1] . Hopkins, Jeffrey L. (2013). Using Commercial Amateur
           Astronomical Spectrographs, p. 257, Springer Science & Business
           Media, ISBN 9783319014425

        .. [2] . Palle, Pere L., Esteban, Cesar. (2014). Asteroseismology, p.
           185, Cambridge University Press, ISBN 9781107470620

        .. [3] . Theveny, Pierre-Michel. (10 September 2001). "Date Format"
           The TPtime Handbook. Media Lab.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${round_index}
    ${dropna}
    ${clean}
    ${skiprows}
    ${names}
    ${source_units}
    ${target_units}
    ${tablefmt}
    """
    tsd = convert_index(
        to,
        interval=interval,
        epoch=epoch,
        input_ts=input_ts,
        columns=columns,
        start_date=start_date,
        end_date=end_date,
        round_index=round_index,
        dropna=dropna,
        clean=clean,
        names=names,
        source_units=source_units,
        target_units=target_units,
        skiprows=skiprows,
    )
    tsutils.printiso(tsd, tablefmt=tablefmt)


# @typic.al
@tsutils.copy_doc(convert_index_cli)
def convert_index(
    to: Literal["datetime", "number"],
    interval=None,
    epoch: Union[
        Literal[
            "julian",
            "reduced",
            "modified",
            "truncated",
            "dublin",
            "cnes",
            "ccsds",
            "lop",
            "lilian",
            "rata_die",
            "mars_sol",
            "unix",
        ],
        pd._typing.TimestampConvertibleTypes,
    ] = "julian",
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    clean=False,
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
):
    """Convert datetime to/from Julian dates from different epochs."""
    # Clip to start_date/end_date if possible.
    if to == "datetime":
        index_type = "number"
        nstart_date = None
        nend_date = None
        nround_index = None
    elif to == "number":
        index_type = "datetime"
        nstart_date = start_date
        nend_date = end_date
        nround_index = round_index
    else:
        raise ValueError(
            tsutils.error_wrapper(
                f"""
The "to" option can be one of "datetime" or "number" not "{to}"."""
            )
        )

    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=nstart_date,
        end_date=nend_date,
        pick=columns,
        round_index=nround_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    allowed = {
        "julian": lambda x: x,
        "reduced": lambda x: x - 2400000,
        "modified": lambda x: x - 2400000.5,
        "truncated": lambda x: np.floor(x - 2440000.5),
        "dublin": lambda x: x - 2415020,
        "cnes": lambda x: x - 2433282.5,
        "ccsds": lambda x: x - 2436204.5,
        "lop": lambda x: x - 2448622.5,
        "lilian": lambda x: np.floor(x - 2299159.5),
        "rata_die": lambda x: np.floor(x - 1721424.5),
        "mars_sol": lambda x: (x - 2405522) / 1.02749,
        "unix": lambda x: (x - 2440587.5),
    }

    dailies = [
        "julian",
        "reduced",
        "modified",
        "truncated",
        "dublin",
        "cnes",
        "ccsds",
        "lop",
        "lilian",
        "rata_die",
        "mars_sol",
    ]

    epoch_dates = {
        "julian": "julian",
        "reduced": "1858-11-16T12",
        "modified": "1858-11-17T00",
        "truncated": "1968-05-24T00",
        "dublin": "1899-12-31T12",
        "cnes": "1950-01-01T00",
        "ccsds": "1958-01-01T00",
        "lop": "1992-01-01T00",
        "lilian": "1582-10-15T00",
        "rata_die": "0001-01-01T00",
        "mars_sol": "1873-12-29T12",
        "unix": "1970-01-01T00",
    }

    if interval is None:
        interval = "D"
    else:
        words = interval.split("-")
        if len(words) == 2:
            warnings.warn(
                """
*
*   The epoch keyword "{}" overrides the anchoring suffix "{}".
*
""".format(
                    epoch, words[1]
                )
            )

            interval = words[0]

    if epoch == "unix" and interval not in ["S", "s"]:
        warnings.warn(
            """
*
*   Typically the unix epoch would has an interval of 'S' (seconds).
*   Instead you gave {}.
*
""".format(
                interval
            )
        )

    if epoch in dailies and interval != "D":
        warnings.warn(
            """
*
*   Typically the {} epoch would has an interval of 'D' (days).
*   Instead you gave {}.
*
""".format(
                epoch, interval
            )
        )

    if to == "number":
        # Index must be datetime - let's make sure
        tsd.index = pd.to_datetime(tsd.index)

        frac = to_offset("D").nanos / to_offset(interval).nanos

        try:
            tsd.index = allowed[epoch](tsd.index.to_julian_date()) * frac
        except KeyError:
            epoch_date = tsutils.parsedate(epoch)
            tsd.index = (
                tsd.index.to_julian_date() - epoch_date.to_julian_date()
            ) * frac

        tsd = tsutils.memory_optimize(tsd)

    elif to == "datetime":
        tsd.index = pd.to_datetime(
            tsd.index.values, origin=epoch_dates.setdefault(epoch, epoch), unit=interval
        )

    if names is None:
        tsd.index.name = "{}_date".format(epoch)

    if to == "datetime":
        index_type = "number"
        nstart_date = start_date
        nend_date = end_date
        nround_index = round_index
    elif to == "number":
        index_type = "datetime"
        nstart_date = None
        nend_date = None
        nround_index = None
    tsd = tsutils.common_kwds(
        tsd, start_date=nstart_date, end_date=nend_date, round_index=nround_index
    )
    return tsd



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter
from scipy.stats import t

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@mando.command("calculate_fdc", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def calculate_fdc_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    percent_point_function=None,
    plotting_position="weibull",
    source_units=None,
    target_units=None,
    sort_values="ascending",
    sort_index="ascending",
    tablefmt="csv",
    add_index=False,
    include_ri=False,
    include_sd=False,
    include_cl=False,
    ci=0.9,
):
    """Return the frequency distribution curve.

    DOES NOT return a time-series.

    Parameters
    ----------
    percent_point_function : str
        [optional, default is None, transformation]

        The distribution used to shift the plotting position values.
        Choose from 'norm', 'lognorm', 'weibull', and None.
    plotting_position : str
        [optional, default is 'weibull', transformation]

        ${plotting_position_table}
    sort_values : str
        [optional, default is 'ascending', input filter]

        Sort order is either 'ascending' or 'descending'.
    sort_index : str
        [optional, default is 'ascending', input filter]

        Sort order is either 'ascending' or 'descending'.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${clean}
    ${tablefmt}
    add_index : bool
        [optional, default is False]

        Add a monotonically increasing index.
    include_ri : bool
        [optional, default is False]

        Include the recurrence interval (sometimes called the return interval).
        This is the inverse of the calculated plotting position defined by the
        equations available with the `plotting_position` keyword.
    include_sd : bool
        [optional, default is False]

        Include a standard deviation column for each column in the
        input.  The equation used is::

            Sd = (Pc(1 - Pc)/N)**0.5

        where::

            Pc is the cumulative probability
            N is the number of values
    include_cl : bool
        [optional, default is False]

        Include two columns showing the upper and lower confidence limit
        for each column in the input.  The equations used are::

            U = Pc + 2(1 - Pc) t Sd
            L = Pc - 2Pc t Sd

        where::

            Pc is the cumulative probability
            t is the Student's "t" value for number of samples and
                confidence interval as defined with `ci` keyword
            Sd is the standard deviation with the equation above
    ci : float
        [optional, default is 0.9]

        This is the confidence interval used when the `include_cl`
        keyword is active.  The confidence interval of 0.9 implies an
        upper limit of 0.95 and a lower limit of 0.05 since 0.9 = 0.95
        - 0.05.
    """
    tsutils.printiso(
        calculate_fdc(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            clean=clean,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            percent_point_function=percent_point_function,
            plotting_position=plotting_position,
            source_units=source_units,
            target_units=target_units,
            sort_values=sort_values,
            sort_index=sort_index,
            add_index=add_index,
            include_ri=include_ri,
            include_sd=include_sd,
            include_cl=include_cl,
            ci=ci,
        ),
        showindex="always",
        tablefmt=tablefmt,
    )


@typic.al
@tsutils.copy_doc(calculate_fdc_cli)
def calculate_fdc(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    percent_point_function=None,
    plotting_position="weibull",
    source_units=None,
    target_units=None,
    sort_values: Literal["ascending", "descending"] = "ascending",
    sort_index: Literal["ascending", "descending"] = "ascending",
    add_index: bool = False,
    include_ri: bool = False,
    include_sd: bool = False,
    include_cl: bool = False,
    ci: tsutils.FloatBetweenZeroAndOne = 0.9,
):
    """Return the frequency distribution curve."""
    sort_values = bool(sort_values == "ascending")

    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    ppf = tsutils.set_ppf(percent_point_function)
    newts = pd.DataFrame()
    for col in tsd:
        tmptsd = tsd[col].dropna()
        if len(tmptsd) > 1:
            return_interval = tsutils.set_plotting_position(
                tmptsd.count(), plotting_position
            )
            xdat = ppf(return_interval)
            tmptsd.sort_values(ascending=sort_values, inplace=True)
            tmptsd.index = xdat * 100
            tmptsd = pd.DataFrame(tmptsd)
            if include_ri is True:
                tmptsd[col + "_ri"] = 1.0 / xdat
            if include_sd is True or include_cl is True:
                sd = (xdat * (1 - xdat) / len(xdat)) ** 0.5
            if include_sd is True:
                tmptsd[col + "_sd"] = sd
            if include_cl is True:
                tval = t.ppf(ci, df=len(xdat) - 1)
                ul = 2 * (1 - xdat) * tval * sd
                ll = 2 * xdat * tval * sd
                tmptsd[col + "_ul"] = (xdat + ul) * 100
                tmptsd[col + "_ll"] = (xdat - ll) * 100
                tmptsd[col + "_vul"] = tmptsd[col] + ul * tmptsd[col]
                tmptsd[col + "_vll"] = tmptsd[col] - ll * tmptsd[col]
        else:
            tmptsd = pd.DataFrame()
        newts = newts.join(tmptsd, how="outer")
    newts.index.name = "Plotting_position"
    newts = newts.groupby(newts.index).first()
    if sort_index == "descending":
        return newts.iloc[::-1]
    if add_index is True:
        newts.reset_index(inplace=True)
    return newts



# -*- coding: utf-8 -*-

import shlex
import subprocess
from unittest import TestCase

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import tstoolbox.tsutils as tsutils
from tstoolbox import tstoolbox

from . import capture


class TestFill(TestCase):
    def setUp(self):
        dindex = pd.date_range("2011-01-01T00:00:00", periods=26, freq="H")
        self.ats = np.ones(26) * 2
        self.ats = pd.DataFrame(
            self.ats,
            index=dindex,
            columns=["Value_with_missing::fill"],
            dtype="Float64",
        )
        self.ats.index.name = "Datetime"

        self.ats_cli = capture.capture(tsutils.printiso, self.ats)

        self.ffill_compare = self.ats.copy()
        self.ffill_compare["Value_with_missing::fill"][
            "2011-01-01T09:00:00":"2011-01-01T12:00:00"
        ] = 3
        self.ffill_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9

        self.ffill_compare_cli = capture.capture(tsutils.printiso, self.ffill_compare)

        self.bfill_compare = self.ats.copy()
        self.bfill_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3
        self.bfill_compare["Value_with_missing::fill"][
            "2011-01-01T10:00:00":"2011-01-01T13:00:00"
        ] = 9

        self.bfill_compare_cli = capture.capture(tsutils.printiso, self.bfill_compare)

        self.linear_compare = self.ats.copy()
        self.linear_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3.0
        self.linear_compare["Value_with_missing::fill"]["2011-01-01T10:00:00"] = 4.5
        self.linear_compare["Value_with_missing::fill"]["2011-01-01T11:00:00"] = 6.0
        self.linear_compare["Value_with_missing::fill"]["2011-01-01T12:00:00"] = 7.5
        self.linear_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9.0

        self.linear_compare_cli = capture.capture(tsutils.printiso, self.linear_compare)

        self.nearest_compare = self.ats.copy()
        self.nearest_compare["Value_with_missing::fill"][
            "2011-01-01T09:00:00":"2011-01-01T11:00:00"
        ] = 3.0
        self.nearest_compare["Value_with_missing::fill"][
            "2011-01-01T12:00:00":"2011-01-01T13:00:00"
        ] = 9.0
        self.nearest_compare = tsutils.memory_optimize(self.nearest_compare)

        self.nearest_compare_cli = capture.capture(
            tsutils.printiso, self.nearest_compare
        )

        self.mean_compare = self.ats.copy()
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T01:00:00"
        ] = 2.4210526315789473
        self.mean_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3.0
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T10:00:00"
        ] = 2.4210526315789473
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T11:00:00"
        ] = 2.4210526315789473
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T12:00:00"
        ] = 2.4210526315789473
        self.mean_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9.0
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T16:00:00"
        ] = 2.4210526315789473
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T22:00:00"
        ] = 2.4210526315789473
        self.mean_compare["Value_with_missing::fill"][
            "2011-01-01T23:00:00"
        ] = 2.4210526315789473

        self.mean_compare_cli = capture.capture(tsutils.printiso, self.mean_compare)

        self.median_compare = self.ats.copy()
        self.median_compare["Value_with_missing::fill"]["2011-01-01T01:00:00"] = 2.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T10:00:00"] = 2.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T11:00:00"] = 2.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T12:00:00"] = 2.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T16:00:00"] = 2.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T22:00:00"] = 2.0
        self.median_compare["Value_with_missing::fill"]["2011-01-01T23:00:00"] = 2.0
        self.median_compare = tsutils.memory_optimize(self.median_compare)

        self.median_compare_cli = capture.capture(tsutils.printiso, self.median_compare)

        self.max_compare = self.ats.copy()
        self.max_compare["Value_with_missing::fill"]["2011-01-01T01:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T10:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T11:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T12:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T16:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T22:00:00"] = 9.0
        self.max_compare["Value_with_missing::fill"]["2011-01-01T23:00:00"] = 9.0
        self.max_compare = tsutils.memory_optimize(self.max_compare)

        self.max_compare_cli = capture.capture(tsutils.printiso, self.max_compare)

        self.min_compare = self.ats.copy()
        self.min_compare["Value_with_missing::fill"]["2011-01-01T01:00:00"] = 2.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T10:00:00"] = 2.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T11:00:00"] = 2.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T12:00:00"] = 2.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T16:00:00"] = 2.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T22:00:00"] = 2.0
        self.min_compare["Value_with_missing::fill"]["2011-01-01T23:00:00"] = 2.0
        self.min_compare = tsutils.memory_optimize(self.min_compare)

        self.min_compare_cli = capture.capture(tsutils.printiso, self.min_compare)

        self.con_compare = self.ats.copy()
        self.con_compare["Value_with_missing::fill"]["2011-01-01T01:00:00"] = 2.42
        self.con_compare["Value_with_missing::fill"]["2011-01-01T09:00:00"] = 3.0
        self.con_compare["Value_with_missing::fill"]["2011-01-01T10:00:00"] = 2.42
        self.con_compare["Value_with_missing::fill"]["2011-01-01T11:00:00"] = 2.42
        self.con_compare["Value_with_missing::fill"]["2011-01-01T12:00:00"] = 2.42
        self.con_compare["Value_with_missing::fill"]["2011-01-01T13:00:00"] = 9.0
        self.con_compare["Value_with_missing::fill"]["2011-01-01T16:00:00"] = 2.42
        self.con_compare["Value_with_missing::fill"]["2011-01-01T22:00:00"] = 2.42
        self.con_compare["Value_with_missing::fill"]["2011-01-01T23:00:00"] = 2.42

        self.con_compare_cli = capture.capture(tsutils.printiso, self.con_compare)

    def test_fill_ffill_direct(self):
        """Test forward fill API."""
        out = tstoolbox.fill(input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.ffill_compare)

    def test_fill_bfill(self):
        """Test backward fill API."""
        out = tstoolbox.fill(method="bfill", input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.bfill_compare)

    # def test_fill_linear(self):
    #     """Test linear interpolation fill API."""
    #     out = tstoolbox.fill(method="linear", input_ts="tests/data_missing.csv")
    #     assert_frame_equal(out, self.linear_compare)

    # def test_fill_nearest(self):
    #     """Test nearest fill API."""
    #     out = tstoolbox.fill(method="nearest", input_ts="tests/data_missing.csv")
    #     assert_frame_equal(out, self.nearest_compare)

    def test_fill_mean(self):
        """Test fill with mean API."""
        out = tstoolbox.fill(method="mean", input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.mean_compare)

    def test_fill_median(self):
        """Test fill with median API."""
        out = tstoolbox.fill(method="median", input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.median_compare)

    def test_fill_max(self):
        """Test fill with max API."""
        out = tstoolbox.fill(method="max", input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.max_compare)

    def test_fill_min(self):
        """Test fill with min API."""
        out = tstoolbox.fill(method="min", input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.min_compare)

    def test_fill_con(self):
        """Test fill with con API."""
        out = tstoolbox.fill(method=2.42, input_ts="tests/data_missing.csv")
        assert_frame_equal(out, self.con_compare)

    @staticmethod
    def test_fill_from():
        """Test fill with from API."""
        out = tstoolbox.fill(
            method="from",
            input_ts="tests/data_fill_from.csv",
            from_columns=[2, 3],
            to_columns=1,
        ).astype("Float64")
        compare = tstoolbox.read("tests/data_fill_from.csv").astype("Float64")
        compare.columns = ["Value::fill", "Value1::fill", "Value2::fill"]
        compare.loc["2000-01-02", "Value::fill"] = 2.5
        compare.loc["2000-01-04", "Value::fill"] = 23.1
        assert_frame_equal(out, compare)

    @staticmethod
    def test_fill_value():
        """Test fill with value API."""
        with pytest.raises(ValueError) as e_info:
            _ = tstoolbox.fill(method="a", input_ts="tests/data_missing.csv")
        assert r"could not convert " in str(e_info.value)

    def test_fill_ffill_cli(self):
        """Test forward fill CLI."""
        args = "tstoolbox fill --input_ts=tests/data_missing.csv"
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate(input=self.ats_cli)[0]
        self.maxDiff = None
        self.assertEqual(out, self.ffill_compare_cli)

    def test_fill_bfill_cli(self):
        """Test backward fill CLI."""
        args = "tstoolbox fill " '--method="bfill" ' "--input_ts=tests/data_missing.csv"
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate(input=self.ats_cli)[0]
        self.maxDiff = None
        self.assertEqual(out, self.bfill_compare_cli)

    # def test_fill_linear_cli(self):
    #     """Test linear fill CLI."""
    #     args = (
    #         "tstoolbox fill "
    #         '--method="linear" '
    #         "--input_ts=tests/data_missing.csv"
    #     )
    #     args = shlex.split(args)
    #     out = subprocess.Popen(
    #         args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
    #     ).communicate(input=self.ats_cli)[0]
    #     self.maxDiff = None
    #     self.assertEqual(out, self.linear_compare_cli)

    # def test_fill_nearest_cli(self):
    #     """Test nearest fill CLI."""
    #     args = (
    #         "tstoolbox fill "
    #         '--method="nearest" '
    #         "--input_ts=tests/data_missing.csv"
    #     )
    #     args = shlex.split(args)
    #     out = subprocess.Popen(
    #         args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
    #     ).communicate(input=self.ats_cli)[0]
    #     self.maxDiff = None
    #     self.assertEqual(out, self.nearest_compare_cli)

    def test_fill_mean_cli(self):
        """Test mean fill CLI."""
        args = "tstoolbox fill " '--method="mean" ' "--input_ts=tests/data_missing.csv"
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate(input=self.ats_cli)[0]
        self.maxDiff = None
        self.assertEqual(out, self.mean_compare_cli)



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import itertools
import os
import warnings
from typing import List, Optional, Tuple

import mando
import numpy as np
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")

ldocstrings = tsutils.docstrings
ldocstrings[
    "xydata"
] = """If the input 'x,y' dataset(s) is organized as
            'index,x1,y1,x2,y2,x3,y3,...,xN,yN' then the 'index' is ignored.
            If there is one 'x,y' dataset then it can be organized as 'index,y'
            where 'index' is used for 'x'.  The "columns" keyword can be used
            to duplicate or change the order of all the data columns."""
ldocstrings[
    "ydata"
] = """Data must be organized as 'index,y1,y2,y3,...,yN'.  The 'index' is
            ignored and all data columns are plotted.  The "columns" keyword
            can be used to duplicate or change the order of all the data
            columns."""
ldocstrings[
    "yone"
] = """Data must be organized as 'index,y1'.  Can only plot one series."""
ldocstrings[
    "secondary_axis"
] = """[optional, default is False]

        * list/tuple: Give the column numbers or names to plot on secondary
          y-axis.
        * (string, string): The first string is the units of the primary axis,
          the second string is the units of the secondary axis if you want just
          unit conversion.  Use any units or combination thereof from the
          "pint" library.
        * (callable, callable): Functions relating relationship between
          primary and secondary axis.  First function will be given the values
          on primary axis and returns valueis on secondary axis.  Second function
          will be do the inverse.  Python API only.
        * string: One of pre-built (callable, callable) combinations.  Can be
          one of "period"."""

MARKER_LIST = [
    ".",
    ",",
    "o",
    "v",
    "^",
    "<",
    ">",
    "1",
    "2",
    "3",
    "4",
    "8",
    "s",
    "p",
    "*",
    "h",
    "H",
    "+",
    "D",
    "d",
    "|",
    "_",
]

LINE_LIST = ["-", "--", "-.", ":", "solid", "dashed", "dashdot", "dotted"]

HATCH_LIST = ["/", "\\", "|", "-", "+", "x", "o", "O", ".", "*"]


def _know_your_limits(xylimits, axis="arithmetic"):
    """Establish axis limits.

    This defines the xlim and ylim as lists rather than strings.
    Might prove useful in the future in a more generic spot.  It
    normalizes the different representations.
    """
    nlim = tsutils.make_list(xylimits, n=2)

    if axis == "normal":
        if nlim is None:
            nlim = [None, None]
        if nlim[0] is None:
            nlim[0] = 0.01
        if nlim[1] is None:
            nlim[1] = 0.99
        if nlim[0] < 0 or nlim[0] > 1 or nlim[1] < 0 or nlim[1] > 1:
            raise ValueError(
                tsutils.error_wrapper(
                    """
Both limits must be between 0 and 1 for the 'normal', 'lognormal', or 'weibull'
axis.

Instead you have {}.
""".format(
                        nlim
                    )
                )
            )

    if nlim is None:
        return nlim

    if nlim[0] is not None and nlim[1] is not None and nlim[0] >= nlim[1]:
        raise ValueError(
            tsutils.error_wrapper(
                """
The second limit must be greater than the first.

You gave {}.
""".format(
                    nlim
                )
            )
        )

    if (
        axis == "log"
        and (nlim[0] is not None and nlim[0] <= 0)
        or (nlim[1] is not None and nlim[1] <= 0)
    ):
        raise ValueError(
            tsutils.error_wrapper(
                """
If log plot cannot have limits less than or equal to 0.

You have {}.
""".format(
                    nlim
                )
            )
        )

    return nlim


@mando.command("plot", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def plot_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    dropna="no",
    index_type="datetime",
    names=None,
    ofilename="plot.png",
    type="time",
    xtitle="",
    ytitle="",
    title="",
    figsize="10,6.0",
    legend=None,
    legend_names=None,
    subplots=False,
    sharex=True,
    sharey=False,
    colors="auto",
    linestyles="auto",
    markerstyles=" ",
    bar_hatchstyles="auto",
    style="auto",
    logx=False,
    logy=False,
    xaxis="arithmetic",
    yaxis="arithmetic",
    xlim=None,
    ylim=None,
    secondary_y=False,
    secondary_x=False,
    mark_right=True,
    scatter_matrix_diagonal="kde",
    bootstrap_size=50,
    bootstrap_samples=500,
    norm_xaxis=False,
    norm_yaxis=False,
    lognorm_xaxis=False,
    lognorm_yaxis=False,
    xy_match_line="",
    grid=False,
    label_rotation=None,
    label_skip=1,
    force_freq=None,
    drawstyle="default",
    por=False,
    invert_xaxis=False,
    invert_yaxis=False,
    round_index=None,
    plotting_position="weibull",
    prob_plot_sort_values="descending",
    source_units=None,
    target_units=None,
    lag_plot_lag=1,
    plot_styles="bright",
    hlines_y=None,
    hlines_xmin=None,
    hlines_xmax=None,
    hlines_colors="auto",
    hlines_linestyles="auto",
    vlines_x=None,
    vlines_ymin=None,
    vlines_ymax=None,
    vlines_colors="auto",
    vlines_linestyles="auto",
):
    r"""Plot data.

    Parameters
    ----------
    ${input_ts}

    ofilename : str, optional
        [optional, defaults to 'plot.png']

        Output filename for the plot.  Extension defines
        the type, for example 'filename.png' will create a PNG file.

        If used within Python, and `ofilename` is None will return the
        Matplotlib figure that can then be changed or added to as
        needed.
    type : {'time', 'xy', 'double_mass', 'bloxplot', 'scatter_matrix',
        'lag_plot', 'autocorrelation', 'bootstrap', 'histogram', 'kde',
        'kde_time', 'bar', 'barh', 'bar_stacked', 'barh_stacked',
        'heatmap', 'norm_xaxis', 'norm_yaxis', 'lognorm_xaxis',
        'lognorm_yaxis', 'weibull_xaxis', 'weibull_yaxis', 'taylor',
        'target'}, optional
        [optional, defaults to 'time']

        The plot type.

        Can be one of the following:

        time
            Standard time series plot.

            Data must be organized as 'index,y1,y2,y3,...,yN'.  The 'index'
            must be a date/time and all data columns are plotted.  Legend names
            are taken from the column names in the first row unless over-ridden
            by the `legend_names` keyword.

        xy
            An 'x,y' plot, also know as a scatter plot.

            ${xydata}
        double_mass
            An 'x,y' plot of the cumulative sum of x and y.

            ${xydata}
        boxplot
            Box extends from lower to upper quartile, with line at the
            median.  Depending on the statistics, the wiskers represent
            the range of the data or 1.5 times the inter-quartile range
            (Q3 - Q1).

            ${ydata}
        scatter_matrix
            Plots all columns against each other in a matrix, with the diagonal
            plots either histogram or KDE probability distribution
            depending on `scatter_matrix_diagonal` keyword.

            ${ydata}
        lag_plot
            Indicates structure in the data.

            ${yone}
        autocorrelation
            Plot autocorrelation.  Only available for a single time-series.

            ${yone}
        bootstrap
            Visually assess aspects of a data set by plotting random
            selections of values.  Only available for a single time-series.

            ${yone}
        histogram
            Calculate and create a histogram plot.  See 'kde' for a smooth
            representation of a histogram.
        kde
            This plot is an estimation of the probability density function
            based on the data called kernel density estimation (KDE).

            ${ydata}
        kde_time
            This plot is an estimation of the probability density function
            based on the data called kernel density estimation (KDE) combined
            with a time-series plot.

            ${ydata}
        bar
            Column plot.
        barh
            A horizontal bar plot.
        bar_stacked
            A stacked column plot.
        barh_stacked
            A horizontal stacked bar plot.
        heatmap
            Create a 2D heatmap of daily data, day of year x-axis, and year for
            y-axis.  Only available for a single, daily time-series.
        norm_xaxis
            Sort, calculate probabilities, and plot data against an
            x axis normal distribution.
        norm_yaxis
            Sort, calculate probabilities, and plot data against an
            y axis normal distribution.
        lognorm_xaxis
            Sort, calculate probabilities, and plot data against an
            x axis lognormal distribution.
        lognorm_yaxis
            Sort, calculate probabilities, and plot data against an
            y axis lognormal distribution.
        weibull_xaxis
            Sort, calculate and plot data against an x axis weibull
            distribution.
        weibull_yaxis
            Sort, calculate and plot data against an y axis weibull
            distribution.
        taylor
            Creates a taylor diagram that compares three goodness of fit
            statistics on one plot.  The three goodness of fit statistics
            calculated and displayed are standard deviation, correlation
            coefficient, and centered root mean square deviation.  The data
            columns have to be organized as
            'observed,simulated1,simulated2,simulated3,...etc.'
        target
            Creates a target diagram that compares three goodness of fit
            statistics on one plot.  The three goodness of fit statistics
            calculated and displayed are bias, root mean square deviation, and
            centered root mean square deviation.  The data columns have to be
            organized as 'observed,simulated1,simulated2,simulated3,...etc.'
    lag_plot_lag : int, optional
        [optional, default to 1]

        The lag used if ``type`` "lag_plot" is chosen.
    xtitle : str
        [optional, default depends on ``type``]

        Title of x-axis.
    ytitle : str
        [optional, default depends on ``type``]

        Title of y-axis.
    title : str
        [optional, defaults to '']

        Title of chart.
    figsize : str
        [optional, defaults to '10,6.5']

        The 'width,height' of plot in inches.
    legend
        [optional, defaults to True]

        Whether to display the legend.
    legend_names : str
        [optional, defaults to None]

        Legend would normally use the time-series names associated with
        the input data.  The 'legend_names' option allows you to
        override the names in the data set.  You must supply a comma
        separated list of strings for each time-series in the data set.
    subplots
        [optional, defaults to False]

        Make separate subplots for each time series.
    sharex
        [optional, default to True]

        In case subplots=True, share x axis.
    sharey
        [optional, default to False]

        In case subplots=True, share y axis.
    colors
        [optional, default is 'auto']

        The default 'auto' will cycle through matplotlib colors in the chosen
        style.

        At the command line supply a comma separated matplotlib
        color codes, or within Python a list of color code strings.

        Can identify colors in four different ways.

        1. Use 'CN' where N is a number from 0 to 9 that gets the Nth color
        from the current style.

        2. Single character code from the table below.

        +------+---------+
        | Code | Color   |
        +======+=========+
        | b    | blue    |
        +------+---------+
        | g    | green   |
        +------+---------+
        | r    | red     |
        +------+---------+
        | c    | cyan    |
        +------+---------+
        | m    | magenta |
        +------+---------+
        | y    | yellow  |
        +------+---------+
        | k    | black   |
        +------+---------+

        3. Number between 0 and 1 that represents the level of gray, where 0 is
        white an 1 is black.

        4. Any of the HTML color names.

        +------------------+
        | HTML Color Names |
        +==================+
        | red              |
        +------------------+
        | burlywood        |
        +------------------+
        | chartreuse       |
        +------------------+
        | ...etc.          |
        +------------------+

        Color reference:
        http://matplotlib.org/api/colors_api.html
    linestyles
        [optional, default to 'auto']

        If 'auto' will iterate through the available matplotlib line types.
        Otherwise on the command line a comma separated list, or a list of
        strings if using the Python API.

        To not display lines use a space (' ') as the linestyle code.

        Separated 'colors', 'linestyles', and 'markerstyles' instead of using
        the 'style' keyword.

        +---------+--------------+
        | Code    | Lines        |
        +=========+==============+
        | ``-``   | solid        |
        +---------+--------------+
        | --      | dashed       |
        +---------+--------------+
        | -.      | dash_dot     |
        +---------+--------------+
        | :       | dotted       |
        +---------+--------------+
        | None    | draw nothing |
        +---------+--------------+
        | ' '     | draw nothing |
        +---------+--------------+
        | ''      | draw nothing |
        +---------+--------------+

        Line reference:
        http://matplotlib.org/api/artist_api.html
    markerstyles
        [optional, default to ' ']

        The default ' ' will not plot a marker.  If 'auto' will iterate through
        the available matplotlib marker types.  Otherwise on the command line
        a comma separated list, or a list of strings if using the Python API.

        Separated 'colors', 'linestyles', and 'markerstyles' instead of using
        the 'style' keyword.

        +-------+----------------+
        | Code  | Markers        |
        +=======+================+
        | .     | point          |
        +-------+----------------+
        | o     | circle         |
        +-------+----------------+
        | v     | triangle down  |
        +-------+----------------+
        | ^     | triangle up    |
        +-------+----------------+
        | <     | triangle left  |
        +-------+----------------+
        | >     | triangle right |
        +-------+----------------+
        | 1     | tri_down       |
        +-------+----------------+
        | 2     | tri_up         |
        +-------+----------------+
        | 3     | tri_left       |
        +-------+----------------+
        | 4     | tri_right      |
        +-------+----------------+
        | 8     | octagon        |
        +-------+----------------+
        | s     | square         |
        +-------+----------------+
        | p     | pentagon       |
        +-------+----------------+
        | ``*`` | star           |
        +-------+----------------+
        | h     | hexagon1       |
        +-------+----------------+
        | H     | hexagon2       |
        +-------+----------------+
        | ``+`` | plus           |
        +-------+----------------+
        | x     | x              |
        +-------+----------------+
        | D     | diamond        |
        +-------+----------------+
        | d     | thin diamond   |
        +-------+----------------+
        | _     | hlines_y       |
        +-------+----------------+
        | None  | nothing        |
        +-------+----------------+
        | ' '   | nothing        |
        +-------+----------------+
        | ''    | nothing        |
        +-------+----------------+

        Marker reference:
        http://matplotlib.org/api/markers_api.html
    style
        [optional, default is None]

        Still available, but if None is replaced by 'colors', 'linestyles', and
        'markerstyles' options.  Currently the 'style' option will override the
        others.

        Comma separated matplotlib style strings per time-series.  Just
        combine codes in 'ColorMarkerLine' order, for example 'r*--' is
        a red dashed line with star marker.
    bar_hatchstyles
        [optional, default to "auto", only used if type equal to "bar", "barh",
        "bar_stacked", and "barh_stacked"]

        If 'auto' will iterate through the available matplotlib hatch types.
        Otherwise on the command line a comma separated list, or a list of
        strings if using the Python API.

        +-----------------+-------------------+
        | bar_hatchstyles | Description       |
        +=================+===================+
        | /               | diagonal hatching |
        +-----------------+-------------------+
        | ``\``           | back diagonal     |
        +-----------------+-------------------+
        | ``|``           | vertical          |
        +-----------------+-------------------+
        | -               | horizontal        |
        +-----------------+-------------------+
        | +               | crossed           |
        +-----------------+-------------------+
        | x               | crossed diagonal  |
        +-----------------+-------------------+
        | o               | small circle      |
        +-----------------+-------------------+
        | O               | large circle      |
        +-----------------+-------------------+
        | .               | dots              |
        +-----------------+-------------------+
        | *               | stars             |
        +-----------------+-------------------+
    logx
        DEPRECATED: use '--xaxis="log"' instead.
    logy
        DEPRECATED: use '--yaxis="log"' instead.
    xlim
        [optional, default is based on range of x values]

        Comma separated lower and upper limits for the x-axis of the
        plot.  For example, '--xlim 1,1000' would limit the plot from
        1 to 1000, where '--xlim ,1000' would base the lower limit on
        the data and set the upper limit to 1000.
    ylim
        [optional, default is based on range of y values]

        Comma separated lower and upper limits for the y-axis of the
        plot.  See `xlim` for examples.
    xaxis : str
        [optional, default is 'arithmetic']

        Defines the type of the xaxis.  One of 'arithmetic', 'log'.
    yaxis : str
        [optional, default is 'arithmetic']

        Defines the type of the yaxis.  One of 'arithmetic', 'log'.
    secondary_y
        ${secondary_axis}
    secondary_x
        ${secondary_axis}
    mark_right
        [optional, default is True]

        When using a secondary_y axis, should the legend label the axis of the
        various time-series automatically.
    scatter_matrix_diagonal : str
        [optional, defaults to 'kde']

        If plot type is 'scatter_matrix', this specifies the plot along the
        diagonal.  One of 'kde' for Kernel Density Estimation or 'hist'
        for a histogram.
    bootstrap_size : int
        [optional, defaults to 50]

        The size of the random subset for 'bootstrap' plot.
    bootstrap_samples
        [optional, defaults to 500]

        The number of random subsets of 'bootstrap_size'.
    norm_xaxis
        DEPRECATED: use '--type="norm_xaxis"' instead.
    norm_yaxis
        DEPRECATED: use '--type="norm_yaxis"' instead.
    lognorm_xaxis
        DEPRECATED: use '--type="lognorm_xaxis"' instead.
    lognorm_yaxis
        DEPRECATED: use '--type="lognorm_yaxis"' instead.
    xy_match_line : str
        [optional, defaults is '']

        Will add a match line where x == y. Set to a line style code.
    grid
        [optional, default is False]

        Whether to plot grid lines on the major ticks.
    label_rotation : int
        [optional]

        Rotation for major labels for bar plots.
    label_skip : int
        [optional]

        Skip for major labels for bar plots.
    drawstyle : str
        [optional, default is 'default']

        'default' connects the points with lines. The
        steps variants produce step-plots. 'steps' is equivalent to 'steps-pre'
        and is maintained for backward-compatibility.

        ACCEPTS::

         ['default' | 'steps' | 'steps-pre' | 'steps-mid' | 'steps-post']
    por
        [optional]

        Plot from first good value to last good value.  Strips NANs
        from beginning and end.
    ${force_freq}
    invert_xaxis
        [optional, default is False]

        Invert the x-axis.
    invert_yaxis
        [optional, default is False]

        Invert the y-axis.
    plotting_position : str
        [optional, default is 'weibull']

        ${plotting_position_table}

        Only used for norm_xaxis, norm_yaxis, lognorm_xaxis,
        lognorm_yaxis, weibull_xaxis, and weibull_yaxis.
    prob_plot_sort_values : str
        [optional, default is 'descending']

        How to sort the values for the probability plots.

        Only used for norm_xaxis, norm_yaxis, lognorm_xaxis,
        lognorm_yaxis, weibull_xaxis, and weibull_yaxis.
    ${columns}
    ${start_date}
    ${end_date}
    ${clean}
    ${skiprows}
    ${dropna}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${round_index}
    plot_styles: str
        [optional, default is "default"]

        Set the style of the plot.  One or more of Matplotlib styles "classic",
        "Solarize_Light2", "bmh", "dark_background", "fast", "fivethirtyeight",
        "ggplot", "grayscale", "seaborn", "seaborn-bright",
        "seaborn-colorblind", "seaborn-dark", "seaborn-dark-palette",
        "seaborn-darkgrid", "seaborn-deep", "seaborn-muted",
        "seaborn-notebook", "seaborn-paper", "seaborn-pastel",
        "seaborn-poster", "seaborn-talk", "seaborn-ticks", "seaborn-white",
        "seaborn-whitegrid", "tableau-colorblind10", and

        SciencePlots styles "science", "grid", "ieee", "scatter", "notebook",
        "high-vis", "bright", "vibrant", "muted", and "retro".

        If multiple styles then each over rides some or all of the
        characteristics of the previous.

        Color Blind Appropriate Styles

        The styles "seaborn-colorblind", "tableau-colorblind10", "bright",
        "vibrant", and "muted" are all styles that are setup to be able to be
        distinguished by someone with color blindness.

        Black, White, and Gray Styles

        The "ieee" style is appropriate for black, white, and gray, however the
        "ieee" also will change the chart size to fit in a column of the "IEEE"
        journal.

        The "grayscale" is another style useful for photo-copyable black,
        white, nd gray.

        Matplotlib styles:
            https://matplotlib.org/3.3.1/gallery/style_sheets/style_sheets_reference.html

        SciencePlots styles:
            https://github.com/garrettj403/SciencePlots
    hlines_y:
        [optional, defaults to None]

        Number or list of y values where to place a horizontal line.
    hlines_xmin:
        [optional, defaults to None]

        List of minimum x values to start the horizontal line.  If a list must
        be same length as `hlines_y`.  If a single number will be used as the
        minimum x values for all horizontal lines.  A missing value or None
        will start at the minimum x value for the entire plot.
    hlines_xmax:
        [optional, defaults to None]

        List of maximum x values to end each horizontal line.  If a list must
        be same length as `hlines_y`.  If a single number will be the maximum
        x value for all horizontal lines.  A missing value or None will end at
        the maximum x value for the entire plot.
    hlines_colors:
        [optional, defaults to None]

        List of colors for the horizontal lines.  If a single color then will
        be used as the color for all horizontal lines.  If a list must be same
        length as `hlines_y`.  If None will take from the color pallette in the
        current plot style.
    hlines_linestyles:
        [optional, defaults to None]

        List of linestyles for the horizontal lines.  If a single linestyle
        then will be used as the linestyle for all horizontal lines.  If a list
        must be same length as `hlines_y`.  If None will take for the standard
        linestyles list.
    vlines_x:
        [optional, defaults to None]

        List of x values where to place a vertical line.
    vlines_ymin:
        [optional, defaults to None]

        List of minimum y values to start the vertical line.  If a list must be
        same length as `vlines_x`.  If a single number will be used as the
        minimum x values for all vertical lines.  A missing value or None will
        start at the minimum x value for the entire plot.
    vlines_ymax:
        [optional, defaults to None]

        List of maximum x values to end each vertical line.  If a list must be
        same length as `vlines_x`.  If a single number will be the maximum
        x value for all vertical lines.  A missing value or None will end at
        the maximum x value for the entire plot.
    vlines_colors:
        [optional, defaults to None]

        List of colors for the vertical lines.  If a single color then will be
        used as the color for all vertical lines.  If a list must be same
        length as `vlines_x`.  If None will take from the color pallette in the
        current plot style.
    vlines_linestyles:
        [optional, defaults to None]

        List of linestyles for the vertical lines.  If a single linestyle then
        will be used as the linestyle for all vertical lines.  If a list must
        be same length as `vlines_x`.  If None will take for the standard
        linestyles list.
    """
    plt = plot(
        input_ts=input_ts,
        columns=columns,
        start_date=start_date,
        end_date=end_date,
        clean=clean,
        skiprows=skiprows,
        dropna=dropna,
        index_type=index_type,
        names=names,
        ofilename=ofilename,
        type=type,
        xtitle=xtitle,
        ytitle=ytitle,
        title=title,
        figsize=figsize,
        legend=legend,
        legend_names=legend_names,
        subplots=subplots,
        sharex=sharex,
        sharey=sharey,
        colors=colors,
        linestyles=linestyles,
        markerstyles=markerstyles,
        bar_hatchstyles=bar_hatchstyles,
        style=style,
        logx=logx,
        logy=logy,
        xaxis=xaxis,
        yaxis=yaxis,
        xlim=xlim,
        ylim=ylim,
        secondary_y=secondary_y,
        secondary_x=secondary_x,
        mark_right=mark_right,
        scatter_matrix_diagonal=scatter_matrix_diagonal,
        bootstrap_size=bootstrap_size,
        bootstrap_samples=bootstrap_samples,
        norm_xaxis=norm_xaxis,
        norm_yaxis=norm_yaxis,
        lognorm_xaxis=lognorm_xaxis,
        lognorm_yaxis=lognorm_yaxis,
        xy_match_line=xy_match_line,
        grid=grid,
        label_rotation=label_rotation,
        label_skip=label_skip,
        force_freq=force_freq,
        drawstyle=drawstyle,
        por=por,
        invert_xaxis=invert_xaxis,
        invert_yaxis=invert_yaxis,
        round_index=round_index,
        plotting_position=plotting_position,
        prob_plot_sort_values=prob_plot_sort_values,
        source_units=source_units,
        target_units=target_units,
        lag_plot_lag=lag_plot_lag,
        plot_styles=plot_styles,
        hlines_y=hlines_y,
        hlines_xmin=hlines_xmin,
        hlines_xmax=hlines_xmax,
        hlines_colors=hlines_colors,
        hlines_linestyles=hlines_linestyles,
        vlines_x=vlines_x,
        vlines_ymin=vlines_ymin,
        vlines_ymax=vlines_ymax,
        vlines_colors=vlines_colors,
        vlines_linestyles=vlines_linestyles,
    )


#    linestyles=[str, ["domain", ["auto", None, "", " ", "  "] + LINE_LIST], None],
#    markerstyles=[str, ["domain", ["auto", None, "", " ", "  "] + MARKER_LIST], None],
#    bar_hatchstyles=[str, ["domain", ["auto", None, "", " ", "  "] + HATCH_LIST], None],
#    hlines_linestyles=[
#        str,
#        ["domain", ["auto", None, "", " ", "  "] + LINE_LIST],
#        None,
#    ],
#    vlines_linestyles=[
#        str,
#        ["domain", ["auto", None, "", " ", "  "] + LINE_LIST],
#        None,
#    ],
@tsutils.transform_args(
    xlim=tsutils.make_list,
    ylim=tsutils.make_list,
    legend_names=tsutils.make_list,
    markerstyles=tsutils.make_list,
    colors=tsutils.make_list,
    linestyles=tsutils.make_list,
    bar_hatchstyles=tsutils.make_list,
    style=tsutils.make_list,
    figsize=tsutils.make_list,
    hlines_y=tsutils.make_list,
    hlines_xmin=tsutils.make_list,
    hlines_xmax=tsutils.make_list,
    hlines_colors=tsutils.make_list,
    hlines_linestyles=tsutils.make_list,
    vlines_x=tsutils.make_list,
    vlines_ymin=tsutils.make_list,
    vlines_ymax=tsutils.make_list,
    vlines_colors=tsutils.make_list,
    vlines_linestyles=tsutils.make_list,
    plot_styles=tsutils.make_list,
)
@typic.al
@tsutils.copy_doc(plot_cli)
def plot(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    clean=False,
    skiprows=None,
    dropna="no",
    index_type="datetime",
    names=None,
    ofilename: Optional[str] = "plot.png",
    type: Literal[
        "time",
        "xy",
        "double_mass",
        "boxplot",
        "scatter_matrix",
        "lag_plot",
        "autocorrelation",
        "bootstrap",
        "histogram",
        "kde",
        "kde_time",
        "bar",
        "barh",
        "bar_stacked",
        "barh_stacked",
        "heatmap",
        "norm_xaxis",
        "norm_yaxis",
        "lognorm_yaxis",
        "lognorm_xaxis",
        "weibull_xaxis",
        "weibull_yaxis",
        "taylor",
        "target",
        "probability_density",
    ] = "time",
    xtitle: str = "",
    ytitle: str = "",
    title: str = "",
    figsize: Tuple[float, float] = "10,6.0",
    legend: Optional[bool] = None,
    legend_names: Optional[List[str]] = None,
    subplots: bool = False,
    sharex: bool = True,
    sharey: bool = False,
    colors: Optional[List[Optional[str]]] = "auto",
    linestyles: Optional[List[Optional[str]]] = "auto",
    markerstyles: Optional[List[Optional[str]]] = " ",
    bar_hatchstyles: Optional[List[Optional[str]]] = "auto",
    style: Optional[List[str]] = "auto",
    logx: bool = False,
    logy: bool = False,
    xaxis: Literal["arithmetic", "log"] = "arithmetic",
    yaxis: Literal["arithmetic", "log"] = "arithmetic",
    xlim: Optional[Tuple[Optional[float], Optional[float]]] = None,
    ylim: Optional[Tuple[Optional[float], Optional[float]]] = None,
    secondary_y=False,
    secondary_x=False,
    mark_right: bool = True,
    scatter_matrix_diagonal: Literal["kde", "hist"] = "kde",
    bootstrap_size: tsutils.IntGreaterEqualToOne = 50,
    bootstrap_samples: tsutils.IntGreaterEqualToOne = 500,
    norm_xaxis: bool = False,
    norm_yaxis: bool = False,
    lognorm_xaxis: bool = False,
    lognorm_yaxis: bool = False,
    xy_match_line: str = "",
    grid: bool = False,
    label_rotation: Optional[float] = None,
    label_skip: tsutils.IntGreaterEqualToOne = 1,
    force_freq: Optional[str] = None,
    drawstyle: str = "default",
    por: bool = False,
    invert_xaxis: bool = False,
    invert_yaxis: bool = False,
    round_index=None,
    plotting_position: Literal[
        "weibull", "benard", "tukey", "gumbel", "hazen", "cunnane", "california"
    ] = "weibull",
    prob_plot_sort_values: Literal["ascending", "descending"] = "descending",
    source_units=None,
    target_units=None,
    lag_plot_lag: tsutils.IntGreaterEqualToOne = 1,
    plot_styles: List[
        Literal[
            "classic",
            "Solarize_Light2",
            "bmh",
            "dark_background",
            "fast",
            "fivethirtyeight",
            "ggplot",
            "grayscale",
            "seaborn",
            "seaborn-bright",
            "seaborn-colorblind",
            "seaborn-dark",
            "seaborn-dark-palette",
            "seaborn-darkgrid",
            "seaborn-deep",
            "seaborn-muted",
            "seaborn-notebook",
            "seaborn-paper",
            "seaborn-pastel",
            "seaborn-poster",
            "seaborn-talk",
            "seaborn-ticks",
            "seaborn-white",
            "seaborn-whitegrid",
            "tableau-colorblind10",
            "science",
            "grid",
            "ieee",
            "scatter",
            "notebook",
            "high-vis",
            "bright",
            "vibrant",
            "muted",
            "retro",
        ]
    ] = "bright",
    hlines_y: Optional[List[float]] = None,
    hlines_xmin: Optional[List[float]] = None,
    hlines_xmax: Optional[List[float]] = None,
    hlines_colors: List[str] = "auto",
    hlines_linestyles: List[Optional[str]] = "auto",
    vlines_x: Optional[List[float]] = None,
    vlines_ymin: Optional[List[float]] = None,
    vlines_ymax: Optional[List[float]] = None,
    vlines_colors: List[str] = "auto",
    vlines_linestyles: List[Optional[str]] = "auto",
):
    r"""Plot data."""
    # Need to work around some old option defaults with the implementation of
    # mando
    legend = bool(legend == "" or legend == "True" or legend is None)

    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    from matplotlib.ticker import FixedLocator
    from plottoolbox import plottoolbox

    tsd = tsutils.common_kwds(
        input_tsd=input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
        por=por,
    )

    if (
        type in ["bootstrap", "heatmap", "autocorrelation", "lag_plot"]
        and len(tsd.columns) != 1
    ):
        raise ValueError(
            tsutils.error_wrapper(
                """
The '{1}' plot can only work with 1 time-series in the DataFrame.
The DataFrame that you supplied has {0} time-series.
""".format(
                    len(tsd.columns), type
                )
            )
        )

    # This is to help pretty print the frequency
    try:
        try:
            pltfreq = str(tsd.index.freq, "utf-8").lower()
        except TypeError:
            pltfreq = str(tsd.index.freq).lower()
        if pltfreq.split(" ")[0][1:] == "1":
            beginstr = 3
        else:
            beginstr = 1
        if pltfreq == "none":
            short_freq = ""
        else:
            # short freq string (day) OR (2 day)
            short_freq = "({})".format(pltfreq[beginstr:-1])
    except AttributeError:
        short_freq = ""

    if legend_names:
        lnames = legend_names
        if len(lnames) != len(set(lnames)):
            raise ValueError(
                tsutils.error_wrapper(
                    """
Each name in legend_names must be unique.
"""
                )
            )
        if len(tsd.columns) == len(lnames):
            renamedict = dict(list(zip(tsd.columns, lnames)))
        elif type in ["xy", "double_mass"] and (
            len(tsd.columns) // 2 == len(lnames) or len(tsd.columns) == 1
        ):
            renamedict = dict(list(zip(tsd.columns[2::2], lnames[1:])))
            renamedict[tsd.columns[1]] = lnames[0]
        else:
            raise ValueError(
                tsutils.error_wrapper(
                    """
For 'legend_names' and most plot types you must have the same number of comma
separated names as columns in the input data.  The input data has {} where the
number of 'legend_names' is {}.

If `type` is 'xy' or 'double_mass' you need to have legend names as
l1,l2,l3,...  where l1 is the legend for x1,y1, l2 is the legend for x2,y2,
...etc.
""".format(
                        len(tsd.columns), len(lnames)
                    )
                )
            )
        tsd.rename(columns=renamedict, inplace=True)
    else:
        lnames = tsd.columns

    if colors is not None and "auto" in colors:
        colors = None

    if linestyles is None:
        linestyles = " "
    elif "auto" in linestyles:
        linestyles = LINE_LIST

    if bar_hatchstyles is None:
        bar_hatchstyles = " "
    elif "auto" in bar_hatchstyles:
        bar_hatchstyles = HATCH_LIST

    if markerstyles is None:
        markerstyles = " "
    elif "auto" in markerstyles:
        markerstyles = MARKER_LIST

    if "auto" not in style:

        nstyle = style
        if len(nstyle) != len(tsd.columns):
            raise ValueError(
                tsutils.error_wrapper(
                    """
You have to have the same number of style strings as time-series to plot.
You supplied '{}' for style which has {} style strings,
but you have {} time-series.
""".format(
                        style, len(nstyle), len(tsd.columns)
                    )
                )
            )
        colors = []
        markerstyles = []
        linestyles = []
        for st in nstyle:
            colors.append(st[0])
            if len(st) == 1:
                markerstyles.append(" ")
                linestyles.append("-")
                continue
            if st[1] in MARKER_LIST:
                markerstyles.append(st[1])
                try:
                    linestyles.append(st[2:])
                except IndexError:
                    linestyles.append(" ")
            else:
                markerstyles.append(" ")
                linestyles.append(st[1:])
    if linestyles is None:
        linestyles = [" "]
    else:
        linestyles = [" " if i in ["  ", None] else i for i in linestyles]
    markerstyles = [" " if i is None else i for i in markerstyles]

    if colors is not None:
        icolors = itertools.cycle(colors)
    else:
        icolors = None
    imarkerstyles = itertools.cycle(markerstyles)
    ilinestyles = itertools.cycle(linestyles)

    # Only for bar, barh, bar_stacked, and barh_stacked.
    ibar_hatchstyles = itertools.cycle(bar_hatchstyles)

    if (
        logx is True
        or logy is True
        or norm_xaxis is True
        or norm_yaxis is True
        or lognorm_xaxis is True
        or lognorm_yaxis is True
    ):
        warnings.warn(
            """
*
*   The --logx, --logy, --norm_xaxis, --norm_yaxis, --lognorm_xaxis, and
*   --lognorm_yaxis options are deprecated.
*
*   For --logx use --xaxis="log"
*   For --logy use --yaxis="log"
*   For --norm_xaxis use --type="norm_xaxis"
*   For --norm_yaxis use --type="norm_yaxis"
*   For --lognorm_xaxis use --type="lognorm_xaxis"
*   For --lognorm_yaxis use --type="lognorm_yaxis"
*
"""
        )

    if xaxis == "log":
        logx = True
    if yaxis == "log":
        logy = True

    if type in ["norm_xaxis", "lognorm_xaxis", "weibull_xaxis"]:
        xaxis = "normal"
        if logx is True:
            logx = False
            warnings.warn(
                """
*
*   The --type={1} cannot also have the xaxis set to {0}.
*   The {0} setting for xaxis is ignored.
*
""".format(
                    xaxis, type
                )
            )

    if type in ["norm_yaxis", "lognorm_yaxis", "weibull_yaxis"]:
        yaxis = "normal"
        if logy is True:
            logy = False
            warnings.warn(
                tsutils.error_wrapper(
                    """
The --type={1} cannot also have the yaxis set to {0}.
The {0} setting for yaxis is ignored.
""".format(
                        yaxis, type
                    )
                )
            )

    xlim = _know_your_limits(xlim, axis=xaxis)
    ylim = _know_your_limits(ylim, axis=yaxis)
    plot_styles = plot_styles + ["no-latex"]
    style_loc = os.path.join(
        os.path.dirname(__file__), os.pardir, "SciencePlots_styles"
    )
    plot_styles = [
        os.path.join(style_loc, i + ".mplstyle")
        if os.path.exists(os.path.join(style_loc, i + ".mplstyle"))
        else i
        for i in plot_styles
    ]
    plt.style.use(plot_styles)

    _, ax = plt.subplots(figsize=figsize)

    if not isinstance(tsd.index, pd.DatetimeIndex) and type == "time":
        raise ValueError(
            tsutils.error_wrapper(
                """
The index is not a datetime index and cannot be plotted as a time-series.
Instead of `type="time"` you might want `type="xy"` or change the index to
a datetime index.
"""
            )
        )

    if type in ["xy", "double_mass"]:
        if tsd.shape[1] > 1 and tsd.shape[1] % 2 != 0:
            raise AttributeError(
                tsutils.error_wrapper(
                    """
The 'xy' and 'double_mass' types must have an even number of columns arranged
as x,y pairs or an x-index and one y data column.  You supplied {} columns.
""".format(
                        tsd.shape[1]
                    )
                )
            )
        colcnt = tsd.shape[1] // 2
    elif type in [
        "norm_xaxis",
        "norm_yaxis",
        "lognorm_xaxis",
        "lognorm_yaxis",
        "weibull_xaxis",
        "weibull_yaxis",
    ]:
        colcnt = tsd.shape[1]

    if type in [
        "xy",
        "double_mass",
        "norm_xaxis",
        "norm_yaxis",
        "lognorm_xaxis",
        "lognorm_yaxis",
        "weibull_xaxis",
        "weibull_yaxis",
    ]:
        plotdict = {
            (False, True): ax.semilogy,
            (True, False): ax.semilogx,
            (True, True): ax.loglog,
            (False, False): ax.plot,
        }

    if type == "time":
        plottoolbox.time(
            tsd,
            legend=legend,
            subplots=subplots,
            sharex=sharex,
            sharey=sharey,
            colors=colors,
            markerstyles=markerstyles,
            linestyles=linestyles,
            logx=logx,
            logy=logy,
            xlim=xlim,
            ylim=ylim,
            mark_right=mark_right,
            figsize=figsize,
            drawstyle=drawstyle,
        )
        for index, line in enumerate(ax.lines):
            if icolors is not None:
                c = next(icolors)
            else:
                c = None
            if imarkerstyles is not None:
                m = next(imarkerstyles)
            else:
                m = None
            if ilinestyles is not None:
                l = next(ilinestyles)
            else:
                l = None
            if c is not None:
                plt.setp(line, color=c)
            plt.setp(line, marker=m)
            plt.setp(line, linestyle=l)
        xtitle = xtitle or "Time"
        if legend is True:
            plt.legend(loc="best")
    elif type in ["taylor"]:
        from ..skill_metrics import centered_rms_dev, taylor_diagram

        ref = tsd.iloc[:, 0]
        std = [np.std(ref)]
        ccoef = [1.0]
        crmsd = [0.0]
        for col in range(1, len(tsd.columns)):
            std.append(np.std(tsd.iloc[:, col]))
            ccoef.append(np.corrcoef(tsd.iloc[:, col], ref)[0][1])
            crmsd.append(centered_rms_dev(tsd.iloc[:, col].values, ref.values))
        taylor_diagram(np.array(std), np.array(crmsd), np.array(ccoef))
    elif type in ["target"]:
        from ..skill_metrics import bias, centered_rms_dev, rmsd, target_diagram

        biases = []
        rmsds = []
        crmsds = []
        ref = tsd.iloc[:, 0].values
        for col in range(1, len(tsd.columns)):
            biases.append(bias(tsd.iloc[:, col].values, ref))
            crmsds.append(centered_rms_dev(tsd.iloc[:, col].values, ref))
            rmsds.append(rmsd(tsd.iloc[:, col].values, ref))
        target_diagram(np.array(biases), np.array(crmsds), np.array(rmsds))
    elif type in ["xy", "double_mass"]:
        # PANDAS was not doing the right thing with xy plots
        # if you wanted lines between markers.
        # Fell back to using raw matplotlib.
        # Boy I do not like matplotlib.

        for colindex in range(colcnt):
            if colcnt == 0:
                ndf = tsd.reset_index()
            else:
                ndf = tsd.iloc[:, colindex * 2 : colindex * 2 + 2]

            ndf = ndf.dropna()

            if type == "double_mass":
                ndf.iloc[:, 0] = ndf.iloc[:, 0].cumsum()
                ndf.iloc[:, 1] = ndf.iloc[:, 1].cumsum()

            oxdata = np.squeeze(ndf.iloc[:, 0].values)
            oydata = np.squeeze(ndf.iloc[:, 1].values)

            if icolors is not None:
                c = next(icolors)
            else:
                c = None

            plotdict[(logx, logy)](
                oxdata,
                oydata,
                linestyle=next(ilinestyles),
                color=c,
                marker=next(imarkerstyles),
                label=lnames[colindex],
                drawstyle=drawstyle,
            )

        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

        if legend is True:
            ax.legend(loc="best")

        if type == "double_mass":
            xtitle = xtitle or "Cumulative {}".format(tsd.columns[0])
            ytitle = ytitle or "Cumulative {}".format(tsd.columns[1])

    elif type in [
        "norm_xaxis",
        "norm_yaxis",
        "lognorm_xaxis",
        "lognorm_yaxis",
        "weibull_xaxis",
        "weibull_yaxis",
    ]:
        ppf = tsutils.set_ppf(type.split("_")[0])
        ys = tsd.iloc[:, :]

        for colindex in range(colcnt):
            oydata = np.array(ys.iloc[:, colindex].dropna())
            if prob_plot_sort_values == "ascending":
                oydata = np.sort(oydata)
            elif prob_plot_sort_values == "descending":
                oydata = np.sort(oydata)[::-1]
            n = len(oydata)

            norm_axis = ax.xaxis
            oxdata = ppf(tsutils.set_plotting_position(n, plotting_position))
            if type in ["norm_yaxis", "lognorm_yaxis", "weibull_yaxis"]:
                oxdata, oydata = oydata, oxdata
                norm_axis = ax.yaxis

            if icolors is not None:
                c = next(icolors)
            else:
                c = None
            plotdict[(logx, logy)](
                oxdata,
                oydata,
                linestyle=next(ilinestyles),
                color=c,
                marker=next(imarkerstyles),
                label=lnames[colindex],
                drawstyle=drawstyle,
            )

        # Make it pretty
        xtmaj = np.array([0.01, 0.1, 0.5, 0.9, 0.99])
        xtmaj_str = ["1", "10", "50", "90", "99"]
        xtmin = np.concatenate(
            [
                np.linspace(0.001, 0.01, 10),
                np.linspace(0.01, 0.1, 10),
                np.linspace(0.1, 0.9, 9),
                np.linspace(0.9, 0.99, 10),
                np.linspace(0.99, 0.999, 10),
            ]
        )
        xtmaj = ppf(xtmaj)
        xtmin = ppf(xtmin)

        norm_axis.set_major_locator(FixedLocator(xtmaj))
        norm_axis.set_minor_locator(FixedLocator(xtmin))

        if type in ["norm_xaxis", "lognorm_xaxis", "weibull_xaxis"]:
            ax.set_xticklabels(xtmaj_str)
            ax.set_ylim(ylim)
            ax.set_xlim(ppf(xlim))

        elif type in ["norm_yaxis", "lognorm_yaxis", "weibull_yaxis"]:
            ax.set_yticklabels(xtmaj_str)
            ax.set_xlim(xlim)
            ax.set_ylim(ppf(ylim))

        if type in ["norm_xaxis", "norm_yaxis"]:
            xtitle = xtitle or "Normal Distribution"
            ytitle = ytitle or tsd.columns[0]
        elif type in ["lognorm_xaxis", "lognorm_yaxis"]:
            xtitle = xtitle or "Log Normal Distribution"
            ytitle = ytitle or tsd.columns[0]
        elif type in ["weibull_xaxis", "weibull_yaxis"]:
            xtitle = xtitle or "Weibull Distribution"
            ytitle = ytitle or tsd.columns[0]

        if type in ["norm_yaxis", "lognorm_yaxis", "weibull_yaxis"]:
            xtitle, ytitle = ytitle, xtitle

        if legend is True:
            ax.legend(loc="best")

    elif type in ["kde", "probability_density"]:
        ax = tsd.plot.kde(
            legend=legend,
            subplots=subplots,
            sharex=sharex,
            sharey=sharey,
            style=None,
            logx=logx,
            logy=logy,
            xlim=xlim,
            ylim=ylim,
            figsize=figsize,
        )
        for index, line in enumerate(ax.lines):
            if icolors is not None:
                c = next(icolors)
            else:
                c = None
            if imarkerstyles is not None:
                m = next(imarkerstyles)
            else:
                m = None
            if ilinestyles is not None:
                l = next(ilinestyles)
            else:
                l = None
            if c is not None:
                plt.setp(line, color=c)
            plt.setp(line, marker=m)
            plt.setp(line, linestyle=l)
        ytitle = ytitle or "Density"
        if legend is True:
            plt.legend(loc="best")
    elif type == "kde_time":
        from scipy.stats.kde import gaussian_kde

        _, (ax0, ax1) = plt.subplots(
            nrows=1,
            ncols=2,
            sharey=True,
            figsize=figsize,
            gridspec_kw={"width_ratios": [1, 4]},
        )
        tsd.plot(
            legend=legend,
            subplots=subplots,
            sharex=sharex,
            sharey=sharey,
            style=None,
            logx=logx,
            logy=logy,
            xlim=xlim,
            ylim=ylim,
            mark_right=mark_right,
            figsize=figsize,
            drawstyle=drawstyle,
            ax=ax1,
        )
        for index, line in enumerate(ax1.lines):
            if icolors is not None:
                c = next(icolors)
            else:
                c = None
            if imarkerstyles is not None:
                m = next(imarkerstyles)
            else:
                m = None
            if ilinestyles is not None:
                l = next(ilinestyles)
            else:
                l = None
            if c is not None:
                plt.setp(line, color=c)
            plt.setp(line, marker=m)
            plt.setp(line, linestyle=l)
        xtitle = xtitle or "Time"
        ylimits = ax1.get_ylim()
        ny = np.linspace(ylimits[0], ylimits[1], 1000)

        # reset to beginning of iterator
        if icolors is not None:
            icolors = itertools.cycle(colors)
        else:
            icolors = None
        imarkerstyles = itertools.cycle(markerstyles)
        ilinestyles = itertools.cycle(linestyles)
        for col, item in enumerate(tsd.columns):
            xvals = tsd.iloc[:, col].dropna().values
            pdf = gaussian_kde(xvals)
            if icolors is not None:
                c = next(icolors)
            ax0.plot(
                pdf(ny),
                ny,
                linestyle=next(ilinestyles),
                color=c,
                marker=next(imarkerstyles),
                label=item,
                drawstyle=drawstyle,
            )
        ax0.set(xlabel="Probability Density", ylabel=ytitle)
    elif type == "boxplot":
        tsd.boxplot(figsize=figsize)
    elif type == "scatter_matrix":
        from pandas.plotting import scatter_matrix

        tsd = tsd.dropna()

        if scatter_matrix_diagonal == "probablity_density":
            scatter_matrix_diagonal = "kde"
        scatter_matrix(tsd, diagonal=scatter_matrix_diagonal, figsize=figsize)
    elif type == "lag_plot":
        from pandas.plotting import lag_plot

        lag_plot(tsd.dropna(), lag=lag_plot_lag, ax=ax)
        xtitle = xtitle or "y(t)"
        ytitle = ytitle or "y(t+{})".format(short_freq or 1)
    elif type == "autocorrelation":
        from pandas.plotting import autocorrelation_plot

        autocorrelation_plot(tsd.dropna(), ax=ax)
        xtitle = xtitle or "Time Lag {}".format(short_freq)
    elif type == "bootstrap":
        from pandas.plotting import bootstrap_plot

        bootstrap_plot(
            tsd, size=bootstrap_size, samples=bootstrap_samples, color="gray"
        )
    elif type == "heatmap":
        # Find beginning and end years
        byear = tsd.index[0].year
        eyear = tsd.index[-1].year
        tsd = tsutils.asbestfreq(tsd)
        if tsd.index.freqstr != "D":
            raise ValueError(
                tsutils.error_wrapper(
                    """
The "heatmap" plot type can only work with daily time series.
"""
                )
            )
        dr = pd.date_range("{}-01-01".format(byear), "{}-12-31".format(eyear), freq="D")
        ntsd = tsd.reindex(index=dr).astype(float)
        groups = ntsd.iloc[:, 0].groupby(pd.Grouper(freq="A"))
        years = pd.DataFrame()
        for name, group in groups:
            ngroup = group.values
            if len(group.values) == 365:
                ngroup = np.append(group.values, [np.nan])
            years[name.year] = ngroup
        years = years.T
        plt.imshow(years, interpolation=None, aspect="auto")
        plt.colorbar()
        yticks = list(range(byear, eyear + 1))
        skip = len(yticks) // 20 + 1
        plt.yticks(range(0, len(yticks), skip), yticks[::skip])
        mnths = [0, 30, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
        mnths_labels = [
            "Jan",
            "Feb",
            "Mar",
            "Apr",
            "May",
            "Jun",
            "Jul",
            "Aug",
            "Sep",
            "Oct",
            "Nov",
            "Dec",
        ]
        plt.xticks(mnths, mnths_labels)
        grid = False
    elif type in ("bar", "bar_stacked", "barh", "barh_stacked"):
        stacked = False
        if "stacked" in type:
            stacked = True
        kind = "bar"
        if "barh" in type:
            kind = "barh"
        if icolors is not None:
            c = [next(icolors) for i in range(len(tsd.columns))]
        else:
            c = None
        tsd.plot(
            ax=ax,
            kind=kind,
            legend=legend,
            stacked=stacked,
            logx=logx,
            logy=logy,
            xlim=xlim,
            ylim=ylim,
            linestyle=None,
            color=c,
        )

        hatches = [next(ibar_hatchstyles) for i in range(len(tsd.columns))]
        hatches = "".join(h * len(tsd.index) for h in hatches)
        for patch, hatch in zip(ax.patches, hatches):
            patch.set_hatch(hatch)

        freq = tsutils.asbestfreq(tsd, force_freq=force_freq).index.freqstr
        if freq is not None:
            if "A" in freq:
                endchar = 4
            elif "M" in freq:
                endchar = 7
            elif "D" in freq:
                endchar = 10
            elif "H" in freq:
                endchar = 13
            else:
                endchar = None
            nticklabels = []
            if kind == "bar":
                taxis = ax.xaxis
            else:
                taxis = ax.yaxis
            for index, i in enumerate(taxis.get_majorticklabels()):
                if index % label_skip:
                    nticklabels.append(" ")
                else:
                    nticklabels.append(i.get_text()[:endchar])
            taxis.set_ticklabels(nticklabels)
            plt.setp(taxis.get_majorticklabels(), rotation=label_rotation)
        if legend is True:
            plt.legend(loc="best")
    elif type == "histogram":
        tsd.hist(figsize=figsize, sharey=sharey, sharex=sharex)

    #    if hlines_y is not None:
    #        hlines_y = tsutils.make_list(hlines_y)
    #        hlines_xmin = tsutils.make_list(hlines_xmin)
    #        hlines_xmax = tsutils.make_list(hlines_xmax)
    #        hlines_colors = tsutils.make_list(hlines_colors)
    #        hlines_linestyles = tsutils.make_list(hlines_colors)
    #    if vlines_x is not None:
    #        vlines_x = tsutils.make_list(vlines_x)
    #        vlines_ymin = tsutils.make_list(vlines_ymin)
    #        vlines_ymax = tsutils.make_list(vlines_ymax)
    #        vlines_colors = tsutils.make_list(vlines_colors)
    #        vlines_linestyles = tsutils.make_list(vlines_colors)
    if type in [
        "time",
        "xy",
        "bar",
        "bar_stacked",
        "histogram",
        "norm_xaxis",
        "lognorm_xaxis",
        "weibull_xaxis",
        "norm_yaxis",
        "lognorm_yaxis",
        "weibull_yaxis",
    ]:
        if hlines_y is not None:
            if type in ["norm_yaxis", "lognorm_yaxis", "weibull_yaxis"]:
                hlines_y = ppf(hlines_y)
            if hlines_xmin is None:
                hlines_xmin = ax.get_xlim()[0]
            if hlines_xmax is None:
                hlines_xmax = ax.get_xlim()[1]
            if hlines_colors == "auto":
                hlines_colors = icolors
            else:
                hlines_colors = itertools.cycle(hlines_colors)
            if hlines_linestyles == "auto":
                hlines_linestyles = ax._get_lines.prop_cycler
            else:
                hlines_linestyles = itertools.cycle(hlines_linestyles)
            for index, line in enumerate(hlines_y):
                try:
                    val = hlines_xmin[index]
                except KeyError:
                    hlines_xmin.append(hlines_xmin[-1])
                try:
                    val = hlines_xmax[index]
                except KeyError:
                    hlines_xmax.append(hlines_xmax[-1])
            for l, xmin, xmax, c, ls in zip(
                hlines_y, hlines_xmin, hlines_xmax, hlines_colors, hlines_linestyles
            ):
                ax.plot([xmin, l], [xmax, l], colors=c, linestyles=l)
        if vlines_x is not None:
            if type in ["norm_xaxis", "lognorm_xaxis", "weibull_xaxis"]:
                vlines_x = ppf(vlines_x)
            plt.vlines(
                vlines_x,
                vlines_ymin,
                vlines_ymax,
                colors=vlines_colors,
                linestyles=vlines_linestyles,
            )

    if type == "xy" and xy_match_line:
        if isinstance(xy_match_line, str):
            xymsty = xy_match_line
        else:
            xymsty = "g--"
        nxlim = ax.get_xlim()
        nylim = ax.get_ylim()
        maxt = max(nxlim[1], nylim[1])
        mint = min(nxlim[0], nylim[0])
        ax.plot([mint, maxt], [mint, maxt], xymsty, zorder=1)
        ax.set_ylim(nylim)
        ax.set_xlim(nxlim)

    plt.xlabel(xtitle)
    plt.ylabel(ytitle)

    if invert_xaxis is True:
        plt.gca().invert_xaxis()
    if invert_yaxis is True:
        plt.gca().invert_yaxis()

    plt.grid(grid)

    plt.title(title)
    plt.tight_layout()
    if ofilename is not None:
        plt.savefig(ofilename)
    return plt



# -*- coding: utf-8 -*-
import numpy as np

from . import utils


def centered_rms_dev(predicted, reference):
    """
    Calculate the centered root-mean-square difference between two variables.

    The latter is calculated using the formula:

    (E')^2 = sum_(n=1)^N [(p_n - mean(p))(r_n - mean(r))]^2/N

    where p is the predicted values, r is the reference values, and
    N is the total number of values in p & r. Note that p & r must
    have the same number of values.

    Input:
    PREDICTED : predicted field
    REFERENCE : reference field

    Output:
    CRMSDIFF : centered root-mean-square (RMS) difference (E')^2

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Nov 24, 2016
    """
    # Check that dimensions of predicted and reference fields match
    utils.check_arrays(predicted, reference)

    # Calculate means
    pmean = np.mean(predicted)
    rmean = np.mean(reference)

    # Calculate (E')^2
    crmsd = np.square((predicted - pmean) - (reference - rmean))
    crmsd = np.sum(crmsd) / predicted.size
    crmsd = np.sqrt(crmsd)

    return crmsd



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

from typing import List, Optional

import mando
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


@mando.command("expanding_window", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def expanding_window_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    statistic="",
    min_periods=1,
    center=False,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Calculate an expanding window statistic.

    Parameters
    ----------
    statistic : str
        [optional, default is '']

        +-----------+----------------------+
        | statistic | Meaning              |
        +===========+======================+
        | corr      | correlation          |
        +-----------+----------------------+
        | count     | count of real values |
        +-----------+----------------------+
        | cov       | covariance           |
        +-----------+----------------------+
        | kurt      | kurtosis             |
        +-----------+----------------------+
        | max       | maximum              |
        +-----------+----------------------+
        | mean      | mean                 |
        +-----------+----------------------+
        | median    | median               |
        +-----------+----------------------+
        | min       | minimum              |
        +-----------+----------------------+
        | skew      | skew                 |
        +-----------+----------------------+
        | std       | standard deviation   |
        +-----------+----------------------+
        | sum       | sum                  |
        +-----------+----------------------+
        | var       | variance             |
        +-----------+----------------------+
    min_periods : int
        [optional, default is 1]

        Minimum number of observations in window required to have a value
    center : boolean
        [optional, default is False]

        Set the labels at the center of the window.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        expanding_window(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            statistic=statistic,
            min_periods=min_periods,
            center=center,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.transform_args(statistic=tsutils.make_list)
@typic.al
@tsutils.copy_doc(expanding_window_cli)
def expanding_window(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    statistic: Optional[
        List[
            Literal[
                "corr",
                "count",
                "cov",
                "kurt",
                "max",
                "mean",
                "median",
                "min",
                "skew",
                "std",
                "sum",
                "var",
            ]
        ]
    ] = None,
    min_periods: tsutils.IntGreaterEqualToZero = 1,
    center: bool = False,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Calculate an expanding window statistic."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    ntsd = tsd.expanding(min_periods=min_periods, center=center)

    if statistic:
        nntsd = pd.DataFrame()
        for stat in statistic:
            ntsd = eval("ntsd.{}()".format(stat))
            ntsd.columns = [
                tsutils.renamer(i, "expanding.{}".format(stat)) for i in ntsd.columns
            ]
            nntsd = nntsd.join(ntsd, how="outer")
    else:
        nntsd = ntsd

    return tsutils.return_input(print_input, tsd, nntsd)



# -*- coding: utf-8 -*-

from __future__ import print_function

import os
from unittest import TestCase

import pytest
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox

test_sinwave = """Datetime,0::,0::peak,0::valley
2000-01-01 00:00:00,0.0,,
2000-01-01 01:00:00,0.258819045103,,
2000-01-01 02:00:00,0.5,,
2000-01-01 03:00:00,0.707106781187,,
2000-01-01 04:00:00,0.866025403784,,
2000-01-01 05:00:00,0.965925826289,,
2000-01-01 06:00:00,1.0,1.0,
2000-01-01 07:00:00,0.965925826289,,
2000-01-01 08:00:00,0.866025403784,,
2000-01-01 09:00:00,0.707106781187,,
2000-01-01 10:00:00,0.5,,
2000-01-01 11:00:00,0.258819045103,,
2000-01-01 12:00:00,1.22464679915e-16,,
2000-01-01 13:00:00,-0.258819045103,,
2000-01-01 14:00:00,-0.5,,
2000-01-01 15:00:00,-0.707106781187,,
2000-01-01 16:00:00,-0.866025403784,,
2000-01-01 17:00:00,-0.965925826289,,
2000-01-01 18:00:00,-1.0,,-1.0
2000-01-01 19:00:00,-0.965925826289,,
2000-01-01 20:00:00,-0.866025403784,,
2000-01-01 21:00:00,-0.707106781187,,
2000-01-01 22:00:00,-0.5,,
2000-01-01 23:00:00,-0.258819045103,,
"""


class TestFilter(TestCase):
    def setUp(self):
        self.ats = tstoolbox.read(os.path.join("tests", "data_sine.csv"))
        self.ats.index.name = "Datetime"
        self.ats.columns = ["Value"]

        self.flat_3 = self.ats.join(
            tstoolbox.read(os.path.join("tests", "data_filter_flat.csv"))
        )
        self.flat_3.columns = ["Value", "Value::filter"]

        self.hanning = self.ats.join(
            tstoolbox.read(os.path.join("tests", "data_filter_hanning.csv"))
        )
        self.hanning.columns = ["Value", "Value::filter"]

        self.fft_lowpass = self.ats.join(
            tstoolbox.read(os.path.join("tests", "data_filter_fft_lowpass.csv"))
        )
        self.fft_lowpass.columns = ["Value", "Value::filter"]

        self.fft_highpass = self.ats.copy()
        self.fft_highpass.columns = ["Value::filter"]
        self.fft_highpass = self.ats.join(self.fft_highpass)

    def test_filter_flat(self):
        out = tstoolbox.filter(
            "flat",
            "lowpass",
            input_ts="tests/data_sine.csv",
            print_input=True,
            window_len=5,
        )
        self.maxDiff = None
        assert_frame_equal(out, self.flat_3, check_column_type=False)

    def test_filter_hanning(self):
        out = tstoolbox.filter(
            "hanning",
            "lowpass",
            input_ts="tests/data_sine.csv",
            print_input=True,
            window_len=5,
        )
        self.maxDiff = None
        assert_frame_equal(out, self.hanning, check_column_type=False)

    def test_filter_fft_lowpass(self):
        out = tstoolbox.filter(
            "fft",
            "lowpass",
            input_ts="tests/data_sine.csv",
            print_input=True,
            lowpass_cutoff=50,
            window_len=5,
        )
        self.maxDiff = None
        assert_frame_equal(out, self.fft_lowpass, check_column_type=False)

    @staticmethod
    def test_small_window_len():
        out = tstoolbox.filter(
            "flat", "lowpass", input_ts="tests/data_sine.csv", window_len=2
        )
        out1 = tstoolbox.read("tests/data_sine.csv")
        out1.columns = ["Value::filter"]
        # NOp
        assert_frame_equal(out, out1)

    @staticmethod
    def test_large_window_len():
        with pytest.raises(ValueError) as e_info:
            _ = tstoolbox.filter(
                "flat", "lowpass", input_ts="tests/data_sine.csv", window_len=1000
            )
        assert r"Input vector (length=" in str(e_info.value)

    @staticmethod
    def test_filter_type():
        with pytest.raises(ValueError) as e_info:
            _ = tstoolbox.filter("flatter", "lowpass", input_ts="tests/data_sine.csv")
        assert r"Given value <'flatter'> fails constraints:" in str(e_info.value)


#    def test_filter_fft_highpass(self):
#        out = tstoolbox.filter('fft_highpass',
#                               input_ts='tests/data_sine.csv',
#                               print_input=True,
#                               cutoff_period=12)
#        self.maxDiff = None
#        assert_frame_equal(out, self.fft_highpass)
#
#    def test_filter_flat_cli(self):
#        args = "tstoolbox filter flat --input_ts='tests/data_sine.csv' --print_input"
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#            stdout=subprocess.PIPE,
#            stdin=subprocess.PIPE).communicate(input=os.path.join('tests', 'data_sine.csv'))[0]
#        self.maxDiff = None
#        self.assertEqual(out, self.flat_3)
#
#    def test_filter_hanning_cli(self):
#        args = ('tstoolbox filter  hanning '
#                '--input_ts=self.ats, '
#                '--print_input=True)')
#        args = shlex.split(args)
#        out = subprocess.Popen(args,
#            stdout=subprocess.PIPE,
#            stdin=subprocess.PIPE).communicate(input=self.ats_cli)[0]
#        self.maxDiff = None
#        self.assertEqual(out, test_sinwave)



# -*- coding: utf-8 -*-

from unittest import TestCase

import pandas as pd
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class TestConvertTZ(TestCase):
    def setUp(self):
        try:
            self.read_direct = (
                pd.read_csv("tests/data_sunspot_EST.csv", index_col=0, parse_dates=[0])
                .tz_localize("UTC")
                .tz_convert("EST")
            )
        except TypeError:
            self.read_direct = pd.read_csv(
                "tests/data_sunspot_EST.csv", index_col=0, parse_dates=[0]
            ).tz_convert("EST")
        self.read_direct = tsutils.memory_optimize(self.read_direct)

    def test_converttz_from_UTC(self):
        out = tstoolbox.converttz("UTC", "EST", input_ts="tests/data_sunspot.csv")
        out.index.name = "Datetime"
        assert_frame_equal(out, self.read_direct)



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np


def overlay_taylor_diagram_circles(axes, cax, option):
    """
    Overlays circle contours on a Taylor diagram.

    Plots circle contours on a Taylor diagram to indicate root mean square
    (RMS) and standard deviation values.

    INPUTS:
    axes   : data structure containing axes information for Taylor diagram
    cax    : handle for plot axes
    option : data structure containing option values. (See
             GET_TAYLOR_DIAGRAM_OPTIONS for more information.)
    option['colrms']       : RMS grid and tick labels color (Default: green)
    option['rincrms']      : Increment spacing for RMS grid
    option['stylerms']     : Linestyle of the RMS grid
    option['tickrms']      : RMS values to plot gridding circles from
                             observation point
    option['tickRMSangle'] : Angle for RMS tick labels with the observation
                             point (Default: 135 deg.)
    option['wi%    dthrms']     : Line width of the RMS grid

    option['colstd']       : STD grid and tick labels color (Default: black)
    option['rincstd']      : Increment spacing for STD grid
    option['stylestd']     : Linestyle of the STD grid
    option['tickstd']      : STD values to plot gridding circles from origin
    option['tickstdangle'] : Angle for STD tick labels with the observation
                             point (Default: .8)
    option['widthstd']     : Line width of the STD grid

    OUTPUTS:
    None.

    See also GET_TAYLOR_DIAGRAM_OPTIONS

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com
    """
    th = np.arange(0, 2 * np.pi, np.pi / 150)
    xunit = np.cos(th)
    yunit = np.sin(th)

    # now really force points on x/y axes to lie on them exactly
    inds = range(0, len(th), (len(th) - 1) // 4)
    xunit[inds[1:5:2]] = np.zeros(2)
    yunit[inds[0:6:2]] = np.zeros(3)

    # DRAW RMS CIRCLES:
    # ANGLE OF THE TICK LABELS
    if option["tickrmsangle"] > 0:
        tickRMSAngle = option["tickrmsangle"]
    else:
        phi = np.arctan2(option["tickstd"][-1], axes["dx"])
        tickRMSAngle = 180 - np.rad2deg(phi)

    c82 = np.cos(tickRMSAngle * np.pi / 180)
    s82 = np.sin(tickRMSAngle * np.pi / 180)
    radius = np.sqrt(
        axes["dx"] ** 2 + axes["rmax"] ** 2 - 2 * axes["dx"] * axes["rmax"] * xunit
    )

    # Define label format
    labelFormat = "{" + option["rmslabelformat"] + "}"

    for iradius in option["tickrms"]:
        phi = th[np.where(radius >= iradius)]
        phi = phi[0]
        ig = np.where(iradius * np.cos(th) + axes["dx"] <= axes["rmax"] * np.cos(phi))
        hhh = plt.plot(
            xunit[ig] * iradius + axes["dx"],
            yunit[ig] * iradius,
            linestyle=option["stylerms"],
            color=option["colrms"],
            linewidth=option["widthrms"],
        )
        if option["showlabelsrms"] == "on":
            xtextpos = (iradius + option["rincrms"] / 20) * c82 + axes["dx"]
            ytextpos = (iradius + option["rincrms"] / 20) * s82
            plt.text(
                xtextpos,
                ytextpos,
                "  " + labelFormat.format(iradius),
                verticalalignment="baseline",
                color=option["colrms"],
                rotation=tickRMSAngle - 90,
            )

    # DRAW STD CIRCLES:
    # draw radial circles
    for i in option["tickstd"]:
        hhh = plt.plot(
            xunit * i,
            yunit * i,
            linestyle=option["stylestd"],
            color=option["colstd"],
            linewidth=option["widthstd"],
        )
        if option["showlabelsstd"] == "on":
            if option["numberpanels"] == 2:
                if len(np.where(option["tickstd"] == 0)) == 0:
                    plt.text(
                        0,
                        -axes["rinc"] / 20,
                        "0",
                        verticalalignment="top",
                        horizontalalignment="center",
                        color=option["colstd"],
                    )
                plt.text(
                    i,
                    -axes["rinc"] / 20,
                    str(i),
                    verticalalignment="top",
                    horizontalalignment="center",
                    color=option["colstd"],
                )
                plt.text(
                    -i,
                    -axes["rinc"] / 20,
                    str(i),
                    verticalalignment="top",
                    horizontalalignment="center",
                    color=option["colstd"],
                )
            else:
                if len(np.where(option["tickstd"] == 0)) == 0:
                    plt.text(
                        -axes["rinc"] / 20,
                        axes["rinc"] / 20,
                        "0",
                        verticalalignment="center",
                        horizontalalignment="right",
                        color=option["colstd"],
                    )
                plt.text(
                    -axes["rinc"] / 20,
                    i,
                    str(i),
                    verticalalignment="center",
                    horizontalalignment="right",
                    color=option["colstd"],
                )

    hhh[0].set_linestyle("-")  # Make outermost STD circle solid

    # Draw circle for outer boundary
    i = option["axismax"]
    hhh = plt.plot(
        xunit * i,
        yunit * i,
        linestyle=option["stylestd"],
        color=option["colstd"],
        linewidth=option["widthstd"],
    )



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

warnings.filterwarnings("ignore")


@mando.command("stdtozrxp", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def stdtozrxp_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    round_index=None,
    source_units=None,
    target_units=None,
    rexchange=None,
):
    """Print out data to the screen in a WISKI ZRXP format.

    Parameters
    ----------
    rexchange
        [optional, default is None]

        The REXCHANGE ID to be written into the zrxp header.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${round_index}
    """
    tsutils.printiso(
        stdtozrxp(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
            rexchange=rexchange,
        )
    )


@tsutils.copy_doc(stdtozrxp_cli)
def stdtozrxp(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    round_index=None,
    source_units=None,
    target_units=None,
    rexchange=None,
):
    """Print out data to the screen in a WISKI ZRXP format."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    if len(tsd.columns) != 1:
        raise ValueError(
            tsutils.error_wrapper(
                """
The "stdtozrxp" command can only accept a single
'time-series, instead it is seeing {}.
""".format(
                    len(tsd.columns)
                )
            )
        )

    if rexchange:
        print("#REXCHANGE{}|*|".format(rexchange))
    for i in range(len(tsd)):
        print(
            (
                "{0.year:04d}{0.month:02d}{0.day:02d}{0.hour:02d}"
                "{0.minute:02d}{0.second:02d}, {1}"
            ).format(tsd.index[i], tsd[tsd.columns[0]][i])
        )



# -*- coding: utf-8 -*-
import numpy as np

from . import rmsd, utils


def skill_score_murphy(predicted, reference):
    """
    Calculate non-dimensional skill score (SS) using Murphy.

    Calculates the non-dimensional skill score (SS) difference between two
    variables PREDICTED and REFERENCE. The skill score is calculated using
    the formula:

    SS = 1 - RMSE^2/SDEV^2

    where RMSE is the root-mean-squre error between the predicted and
    reference values

    (RMSE)^2 = sum_(n=1)^N (p_n - r_n)^2/N

    and SDEV is the standard deviation of the reference values

    SDEV^2 = sum_(n=1)^N [r_n - mean(r)]^2/(N-1)

    where p is the predicted values, r is the reference values, and
    N is the total number of values in p & r. Note that p & r must
    have the same number of values.

    Input:
    PREDICTED : predicted field
    REFERENCE : reference field

    Output:
    SS : skill score

    Reference:
    Allan H. Murphy, 1988: Skill Scores Based on the Mean Square Error
    and Their Relationships to the Correlation Coefficient. Mon. Wea.
    Rev., 116, 2417-2424.
    doi: http//dx.doi.org/10.1175/1520-0493(1988)<2417:SSBOTM>2.0.CO;2

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 7, 2016
    """
    # Check that dimensions of predicted and reference fields match
    utils.check_arrays(predicted, reference)

    # Calculate RMSE
    rmse2 = rmsd(predicted, reference) ** 2

    # Calculate standard deviation
    sdev2 = np.std(reference, ddof=1) ** 2

    # Calculate skill score
    ss = 1 - rmse2 / sdev2

    return ss



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import mando
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils


@mando.command("date_offset", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def date_offset_cli(
    intervals,
    offset,
    columns=None,
    dropna="no",
    clean=False,
    skiprows=None,
    index_type="datetime",
    names=None,
    input_ts="-",
    start_date=None,
    end_date=None,
    source_units=None,
    target_units=None,
    round_index=None,
    tablefmt="csv",
):
    """Apply a date offset to a time-series index.

    If you want to adjust to a different time-zone then should use the
    "converttz" tstoolbox command.

    Parameters
    ----------
    intervals: int
        Number of intervals of `offset` to shift the time index.  A positive
        integer moves the index forward, negative moves it backwards.
    offset: str
        Any of the Pandas offset codes.  This is only the offset code
        and doesn't include a prefixed interval.

        ${pandas_offset_codes}
    ${input_ts}
    ${start_date}
    ${end_date}
    ${columns}
    ${round_index}
    ${dropna}
    ${clean}
    ${skiprows}
    ${index_type}
    ${source_units}
    ${target_units}
    ${names}
    ${tablefmt}
    """
    tsutils.printiso(
        date_offset(
            intervals,
            offset,
            columns=columns,
            dropna=dropna,
            clean=clean,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            input_ts=input_ts,
            start_date=start_date,
            end_date=end_date,
            source_units=source_units,
            target_units=target_units,
            round_index=round_index,
        ),
        showindex="always",
        tablefmt=tablefmt,
    )


@tsutils.copy_doc(date_offset_cli)
def date_offset(
    intervals: tsutils.IntGreaterEqualToOne,
    offset: str,
    columns=None,
    dropna="no",
    clean=False,
    skiprows=None,
    input_ts="-",
    start_date=None,
    end_date=None,
    names=None,
    index_type="datetime",
    source_units=None,
    target_units=None,
    round_index=None,
):
    """Apply an offset to a time-series."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    return tsd.shift(intervals, offset)
